/*------------------------------------------------------------------------------
 bin/djv_convert/djv_convert.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_base/user.h>
#include <djv_base/file.h>
#include <djv_base/sys.h>
#include <djv_base/box.h>
#include <djv_image/time.h>
#include <djv_image/io.h>
#include <djv_image/app.h>

namespace
{
using namespace djv;

class App : public image::App
{
public:

  App(int, char **);

private:

  void work() throw (Error);

  Str label_time(int, const image::time::Speed &) const;
  Str label_time(double) const;
  Str label_image(const image::Info &, const image::Time &) const;
  
  void cmdln() throw (Error);
  Str help() const;

  File _in, _out;
  image::gl::data::Draw _draw;
  Str _load, _save;
  int _layer, _proxy;
  Str _time_start, _time_end;
  int _timeout;
  std::auto_ptr<image::PIXEL> _pixel;
  std::auto_ptr<image::time::FPS> _speed;
  image::Tag _tag;
  bool _tag_auto;
  bool _timecode;
};

App::App(int argc, char ** argv) :
  image::App("djv_convert", argc, argv),
  _layer(0), _proxy(0), _timeout(0), _tag_auto(true), _timecode(true)
{
  DJV_DBG_OBJ("App::App");
  
  // Command line.

  if (app::EXIT_DEFAULT == exit())
    try { cmdln(); }
    catch (Error in)
    {
      error(str::Format("Command Line: %1").arg(in));
      exit(app::EXIT_ERROR);
      return;
    }
  if (exit() != app::EXIT_DEFAULT) return;

  if (_in.seq_valid()) _in.type = file::SEQ;
  if (_out.seq_valid()) _out.type = file::SEQ;
  
  // Work.

  try { work(); }
  catch (Error error) { this->error(error); exit(app::EXIT_ERROR); }
}

void App::work() throw (Error)
{
  DJV_DBG_OBJ("App::work");
  DJV_DBG_PRN("input = " << _in);
  DJV_DBG_PRN("output = " << _out);

  djv::time::Timer timer;
  timer.start();
  
  // Open input.

  std::auto_ptr<image::io::Load> load;
  image::Info info;
  image::Tag tag;
  image::Time time;

  Error error;
  while (! load.get())
  {
    try
    {
      load = std::auto_ptr<image::io::Load>(image::io::load()->get(
        _in, &info, &tag, &time, _load));
      tag.add(_tag);
    }
    catch (Error in) { error = in; }
    if (! load.get() && _timeout)
    {
      print("Timeout...");
      djv::time::sleep(1);
    } else
      break;
  }
  if (! load.get())
    throw Error(str::Format("%1: %2").arg(error).arg(_in.name()));

  if (_time_start.size())
    time.start = _timecode ?
      image::time::timecode_to_frame(
        image::time::str_to_timecode(_time_start), time.speed) :
      str::str_to_int(_time_start);
  if (_time_end.size())
    time.end = _timecode ?
      image::time::timecode_to_frame(
        image::time::str_to_timecode(_time_end), time.speed) :
      str::str_to_int(_time_end);
  
  print(str::Format("%1 %2").arg(_in).arg(label_image(info, time)));
  
  // Open output.

  image::Info save_info;
  
  if (vec::size_valid(_draw.scale))
    save_info.size = vec::to_int(_draw.scale);
  else
  {
    save_info.size = info.size;
    _draw.scale = vec::to_float(info.size);
  }

  save_info.pixel = _pixel.get() ? *_pixel : info.pixel;
  
  image::Time save_time;
  if (_out.seq_valid())
  {
    save_time.start = save_time.end = _out.seq.start;
    if (_out.seq.start != _out.seq.end)
      save_time.end = _out.seq.end;
  }
  if (save_time.start == save_time.end)
    save_time.end = save_time.start + time.end - time.start;
  save_time.speed = _speed.get() ? *_speed : time.speed;

  std::auto_ptr<image::io::Save> save;
  try
  {
    save = std::auto_ptr<image::io::Save>(
      image::io::save()->get(_out, save_info, tag, save_time, _save));
  }
  catch (Error in)
  {
    throw Error(str::Format("%1: %2").arg(_out.name()).arg(in));
  }

  print(str::Format("%1 %2").arg(_out).arg(label_image(save_info, save_time)));

  // Work.

  const int length = save_time.end - save_time.start + 1;
  const int progress_length = 10;
  int progress_count = 0;
  double frame_time = 0.0;
  for (int i = 0; i < length; ++i)
  {
    djv::time::Timer frame_timer;
    frame_timer.start();
    
    // Load.

    const image::Image * image = 0;
    
    int timeout = _timeout;
    while (! image)
    {
      try { image = load->get(time.start + i, _layer, _proxy); }
      catch (Error in) { error = in; }
      if (! image && timeout > 0)
      {
        //print("Timeout...");
        --timeout;
        djv::time::sleep(1);
      }
      else
        break;
    }
    if (! image)
      throw Error(str::Format("%1: %2").arg(error).arg(_in.name()));
    DJV_DBG_PRN("image = " << *image);

    // Image tags.
    
    tag = image->tag;
    if (_tag_auto)
    {
      tag[image::tag::label()[image::tag::CREATOR]] = user::current();
      tag[image::tag::label()[image::tag::TIME]] =
        djv::time::time_to_str(djv::time::current());
      tag[image::tag::label()[image::tag::TIMECODE]] =
        image::time::timecode_to_str(image::time::frame_to_timecode(
          save_time.start + i, save_time.speed));
    }
    tag.add(_tag);
    
    // Process.

    image::Image tmp(save_info);
    tmp.tag = tag;
    DJV_DBG_PRN("tmp = " << tmp);
    
    _draw.profile = image->profile;
    image::gl::data::copy(*image, &tmp, _draw);
    
    // Save.

    DJV_DBG_PRN("output = " << tmp);
    try
    {
      save->set(tmp, save_time.start + i);
    }
    catch (Error in)
    {
      throw Error(str::Format("%1: %2").arg(_out.name()).arg(in));
    }

    // Statistics.

    frame_timer.check();
    frame_time += frame_timer.seconds();

    if (length > 1)
      if (i / double(length) * progress_length > progress_count)
      {
        print(str::Format("[%1%] Estimated Time = %2").
          arg(int(i / double(length) * 100.0), 3).
          arg(label_time(frame_time / double(i + 1) * (length - (i + 1)))));
        ++progress_count;
      }
  }
  if (length > 1) print("[100%] ", false);
  
  try
  {
    save->close();
  }
  catch (Error in)
  {
    throw Error(str::Format("%1: %2").arg(_out.name()).arg(in));
  }

  timer.check();
  print(str::Format("Time Elapsed = %1 (%2 frames/second)").
    arg(label_time(timer.seconds())).arg(length / timer.seconds(), 0, -2));
}

Str App::label_time(int in, const image::time::Speed & speed) const
{
  return _timecode ?
    image::time::timecode_to_str(image::time::frame_to_timecode(in, speed)) :
    str::int_to_str(in);
}

Str App::label_time(double in) const
{
  int hour, minute;
  double second;
  djv::time::seconds_to_time(in, &hour, &minute, &second);
  return str::Format("%1:%2:%3").
    arg(hour, 2, '0').arg(minute, 2, '0').arg(second, 5, 2, '0');
}

Str App::label_image(const image::Info & in, const image::Time & time) const
{
  return str::Format("%1x%2:%3 %4 %5-%6@%7").
    arg(in.size.x).arg(in.size.y).arg(aspect(in.size), 0, -2).
    arg(image::pixel::label_pixel()[in.pixel]).
    arg(label_time(time.start, time.speed)).
    arg(label_time(time.end, time.speed)).
    arg(image::time::speed_to_float(time.speed));
}

void App::cmdln() throw (Error)
{
  if (! cmdln_list.size()) { exit(app::EXIT_HELP); return; }
  
  List<Str> arg_list;
  while (cmdln_list.size())
  {
    const Str arg = cmdln_list.pop_front();
    
    // Options.
    
    if (str::compare("-x_timecode", arg) || str::compare("-t", arg))
      _timecode = false;
    
    // Image options.

    else if (str::compare("-mirror_x", arg))
      _draw.mirror_x = true;
    else if (str::compare("-mirror_y", arg))
      _draw.mirror_y = true;
    else if (str::compare("-scale", arg))
    {
      if (! vec::serialize(&cmdln_list, &_draw.scale)) throw Error(arg);
    }
    else if (str::compare("-channel", arg))
    {
      if (! str::serialize(&cmdln_list, &_draw.channel,
        image::gl::data::label_channel())) throw Error(arg);
    }
    /*else if (str::compare("-filter", arg))
    {
      if (! str::serialize(&cmdln_list, &_draw.filter,
        image::gl::data::label_filter())) throw Error(arg);
    }*/
    
    // Load options.

    else if (str::compare("-load", arg))
    {
      Str plugin, option;
      if (! str::serialize(&cmdln_list, &plugin)) throw Error(arg);
      if (! str::serialize(&cmdln_list, &option)) throw Error(arg);
      if (! image::io::load()->option(plugin, option, &cmdln_list))
        throw Error(arg);
    }
    else if (str::compare("-load_plugin", arg))
    {
      if (! str::serialize(&cmdln_list, &_load)) throw Error(arg);
    }
    else if (str::compare("-layer", arg))
    {
      if (! str::serialize(&cmdln_list, &_layer)) throw Error(arg);
    }
    else if (str::compare("-proxy", arg))
    {
      if (! str::serialize(&cmdln_list, &_proxy)) throw Error(arg);
    }
    else if (str::compare("-time", arg))
    {
      if (! str::serialize(&cmdln_list, &_time_start)) throw Error(arg);
      if (! str::serialize(&cmdln_list, &_time_end)) throw Error(arg);
    }
    else if (str::compare("-timeout", arg))
    {
      if (! str::serialize(&cmdln_list, &_timeout)) throw Error(arg);
    }
    
    // Save options.
    
    else if (str::compare("-save", arg))
    {
      Str plugin, option;
      if (! str::serialize(&cmdln_list, &plugin)) throw Error(arg);
      if (! str::serialize(&cmdln_list, &option)) throw Error(arg);
      if (! image::io::save()->option(plugin, option, &cmdln_list))
        throw Error(arg);
    }
    else if (str::compare("-save_plugin", arg))
    {
      if (! str::serialize(&cmdln_list, &_save)) throw Error(arg);
    }
    else if (str::compare("-pixel", arg) || str::compare("-p", arg))
    {
      image::PIXEL value = image::PIXEL(0);
      if (! str::serialize(&cmdln_list, &value, image::pixel::label_pixel()))
        throw Error(arg);
      _pixel = std::auto_ptr<image::PIXEL>(new image::PIXEL(value));
    }
    else if (str::compare("-speed", arg))
    {
      image::time::FPS value = image::time::FPS(0);
      if (! str::serialize(&cmdln_list, &value, image::time::label_fps()))
        throw Error(arg);
      _speed = std::auto_ptr<image::time::FPS>(new image::time::FPS(value));
    }
    else if (str::compare("-tag", arg))
    {
      Str name, value;
      if (! str::serialize(&cmdln_list, &name)) throw Error(arg);
      if (! str::serialize(&cmdln_list, &value)) throw Error(arg);
      _tag[name] = value;
    }
    else if (str::compare("-x_tag", arg)) _tag_auto = false;

    // Arguments.
    
    else arg_list += arg;
  }
  
  if (arg_list.size() < 1) throw Error("input");
  _in = arg_list[0];
  if (arg_list.size() < 2) throw Error("output");
  _out = arg_list[1];
  if (arg_list.size() > 2) throw Error(arg_list[2]);
}

const Str _help =
"\n"
"djv_convert\n"
"-----------\n"
"\n"
"Image conversion utility.\n"
"\n"
"Command Line\n"
"------------\n"
"\n"
"Usage:\n"
"\n"
"djv_convert (input) (output) [option]...\n"
"\n"
"Options:\n"
"\n"
"-x_timecode, -t - Disable timecode display.\n"
"\n"
"Image Options:\n"
"\n"
"-mirror_x - Mirror image horizontally.\n"
"-mirror_y - Mirror image vertically.\n"
"-scale (width) (height) - Scale image.\n"
"-channel (value) - Channel: %1.\n"
//"-filter (value) - Filter: %2. Default: %3.\n"
"\n"
"Load Options:\n"
"\n"
"-load (plugin) (option) (value) - Set a plugin option:\n"
"%4"
"\n"
"-load_plugin (value) - Specify load plugin: %5.\n"
"\n"
"-layer (value) - Image layer index.\n"
"\n"
"-proxy (value) - Proxy scale.\n"
"\n"
"-time (start) (end) - Start and end frames.\n"
"-timeout (value) - Maximum number of seconds to wait for each input frame. "
"Default: %6.\n"
"\n"
" Save Options:\n"
"\n"
"-save (plugin) (option) (value) - Set a plugin option:\n"
"%7"
"\n"
"-save_plugin (value) - Specify save plugin: %8.\n"
"\n"
"-pixel, -p (value) - Pixel type: %9.\n"
"\n"
"-speed (value) - Speed: %10.\n"
"\n"
"-tag (name) (value) - Set an image tag.\n"
"-x_tag - Disable automatic image tag generation.\n"
"\n"
"%11"
"\n"
"Examples\n"
"--------\n"
"\n"
"Convert an image:\n"
"> djv_convert input.sgi output.tga\n"
"\n"
"Convert an image sequence:\n"
"> djv_convert input.1-100.sgi output.1.tga\n"
"Only the first frame of the output needs to be specified.\n"
"\n"
"Convert a movie to an image sequence:\n"
"> djv_convert input.mov output.1.tga\n"
"\n"
"Convert an image sequence to a movie:\n"
"> djv_convert input.0001-0100.dpx output.mov\n"
"\n"
"Convert pixel type:\n"
"> djv_convert input.sgi output.sgi -pixel \"rgb u16\"\n"
"\n"
"Scale an image:\n"
"> djv_convert input.sgi output.sgi -scale 2048 1556\n"
"\n"
"Save video lookup table to a file:\n"
"> djv_convert 0.vlut lut.1dl -load_plugin vlut\n"
"\n"
"Load video lookup table from a file:\n"
"> djv_convert lut.1dl 0.vlut -save_plugin vlut\n";

Str App::help() const
{
  return str::Format(_help).
    arg(str::join(image::gl::data::label_channel(), ", ")).
    //arg(str::join(image::gl::data::label_filter(), ", ")).
    //arg(image::gl::data::label_filter()[_options.filter]).
    arg(image::io::load()->help()).
    arg(str::join(image::io::load()->name_list(), ", ")).
    arg(_timeout).
    arg(image::io::save()->help()).
    arg(str::join(image::io::save()->name_list(), ", ")).
    arg(str::join(image::pixel::label_pixel(), ", ")).
    arg(str::join(image::time::label_fps(), ", ")).
    arg(image::App::help());
}

}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = ::App(argc, argv).exec(); }
  catch (djv::Error in) { djv::error::print(in); }
  return r;
}

