/*------------------------------------------------------------------------------
 bin/djv_view/view.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "view.h"
#include <djv_base/math.h>
#include <djv_image/time.h>
#include <djv_ui/style.h>
#include <FL/Fl.H>
#include <FL/gl.h>

namespace djv_view {
namespace view
{

Widget::Widget() :
  _grid(GRID(0)),
  _hud(false),
  _inside(false)
{
  text_font(ui::style::global()->font_bold());
}

void Widget::zoom(double in)
{
  ui::image::Widget::zoom(in, _inside ? _mouse : (geom().size / 2));
}

void Widget::zoom(double in, const Vec2i & focus)
{
  ui::image::Widget::zoom(in, focus);
}

void Widget::grid(GRID in) { _grid = in; }
void Widget::grid_color(const image::Color & in) { _grid_color = in; }

void Widget::hud(bool in) { _hud = in; }
void Widget::hud_info(const Hud_Info & in) { _hud_info = in; }
void Widget::hud_foreground(const image::Color & in) { _hud_foreground = in; }
void Widget::hud_background(const image::Color & in) { _hud_background = in; }

int Widget::handle(int in)
{
  //DJV_DBG_OBJ("view::Widget::handle");
  //DJV_DBG_PRN("in = " << in);
  
  const Vec2i mouse = Vec2i(Fl::event_x(), geom().h - 1 - Fl::event_y());
  switch (in)
  {
    case FL_PUSH:
    {
      take_focus();
      
      _mouse = mouse;
      _mouse_td = _mouse;
      
      if (Fl::event_ctrl())
      {
        _view_tmp = view();
        cursor(FL_CURSOR_MOVE);
      }
      else
      {
        cursor(FL_CURSOR_CROSS);
        pick_sig.emit(mouse);
      }
    } return 1;
    
    case FL_MOVE: { _mouse = mouse; } return 1;
    
    case FL_DRAG:
    {
      DJV_DBG_OBJ("view::Widget::handle");
    
      _mouse = mouse;
      
      if (Fl::event_ctrl())
      {
        view(_view_tmp + (_mouse - _mouse_td));
        cursor(FL_CURSOR_MOVE);
        redraw();
      }
      else
      {
        cursor(FL_CURSOR_CROSS);
        pick_sig.emit(mouse);
      }
    } return 1;
    
    case FL_RELEASE:
      cursor(FL_CURSOR_DEFAULT);
      return 1;
      
    case FL_ENTER: _inside = true; return 1;
    case FL_LEAVE: _inside = false; return 1;
    case FL_FOCUS:
    case FL_UNFOCUS: return 1;
    
    case FL_DND_ENTER:
    case FL_DND_DRAG:
    case FL_DND_LEAVE:
    case FL_DND_RELEASE: return 1;
    case FL_PASTE:
    {
      DJV_DBG_OBJ("view::Widget::handle");
      
      Str tmp(Fl::event_text());
      
      // XXX Convert URL.
      
      static const Str url = "file://";
      if (tmp.find(url) != Str::npos)
        tmp.replace(0, url.size(), "");
      int i = int(tmp.size()) - 1;
      for (; i >= 0; --i)
        if (tmp[i] != '\n' && tmp[i] != '\r') break;
      tmp = Str(tmp, 0, i + 1);
      
      DJV_DBG_PRN("file = " << tmp);
      file_sig.emit(tmp);

    } return 1;
  }
  return ui::image::Widget::handle(in);
}

void Widget::draw()
{
  if (! visible()) return;
  
  DJV_DBG_OBJ("view::Widget::draw");
  
  ui::image::Widget::draw();
  if (_grid) draw_grid();
  if (_hud) draw_hud();
}

void Widget::draw_grid()
{
  DJV_DBG_OBJ("view::Widget::draw_grid");
  DJV_DBG_PRN("size = " << geom().size);
  DJV_DBG_PRN("grid = " << label_grid()[_grid]);
  
  int inc;
  switch (_grid)
  {
    case GRID_1x1: inc = 1; break;
    case GRID_10x10: inc = 10; break;
    case GRID_100x100: inc = 100; break;
    default: break;
  }
  
  // Bail if too small.
  
  if ((inc * zoom()) <= 2) return;
  
  // Compute view area.
  
  const Box2i area = Box2i(
    vec::floor(vec::to_float(-view()) / zoom() / double(inc)) - 1,
    vec::ceil(vec::to_float(geom().size) / zoom() / double(inc)) + 2) * inc;
  DJV_DBG_PRN("area = " << area);
  
  // Draw.
  
  djv::image::Color color(djv::image::pixel::RGB_F32);
  djv::image::color::convert(_grid_color, &color);
  glColor3f(color.get_f32(0), color.get_f32(1), color.get_f32(2));
  
  glPushMatrix();
  glTranslated(view().x, view().y, 0);
  glScaled(zoom(), zoom(), 1.0);
  
  glBegin(GL_LINES);
  for (int y = 0; y <= area.h; y += inc)
  {
    glVertex2i(area.x, area.y + y);
    glVertex2i(area.x + area.w, area.y + y);
  }
  for (int x = 0; x <= area.w; x += inc)
  {
    glVertex2i(area.x + x, area.y);
    glVertex2i(area.x + x, area.y + area.h);
  }
  glEnd();
  
  glPopMatrix();
}

void Widget::draw_hud()
{
  //DJV_DBG_OBJ("view::Widget::draw_hud");
  
  const Box2i geom = this->geom();
  const int margin = ui::style::global()->margin_widget();

  // Setup.
  
  // XXX WTFBBQ? Somehow reset the OpenGL state so FLTK's text drawing
  // functions work?
  
  glGetError();

  //glViewport(0, 0, geom.w, geom.h);
  //image::gl::ortho(geom.w, geom.h);

  gl_font(text_font(), ui::style::global()->font_size());
  
  // Image information.

  int x = margin;
  int y = geom.h - gl_height() - margin;
  Str str = str::Format("File = %1").arg(_hud_info.file);
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
  y -= gl_height();
  
  str = str::Format("Layer = %1").arg(_hud_info.image.name);
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
  y -= gl_height();
  
  str = str::Format("Size = %1x%2:%3").arg(_hud_info.image.size.x).
    arg(_hud_info.image.size.y).arg(vec::aspect(_hud_info.image.size), 0, -2);
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
  y -= gl_height();
  
  str = str::Format("Proxy = %1").arg(_hud_info.image.proxy);
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
  y -= gl_height();
  
  str = str::Format("Pixel = %1").
    arg(image::pixel::label_pixel()[_hud_info.image.pixel]);
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
  
  // Image tags.
 
  List<Str> tag;
  const List<Str> key_list = _hud_info.tag.key_list();
  for (uint i = 0; i < key_list.size(); ++i)
    tag += str::Format("Tag %1 = %2").arg(key_list[i]).
      arg(_hud_info.tag[key_list[i]]);
  
  y = gl_height() * (int(tag.size()) - 1) + margin;
  for (List<Str>::const_iterator i = tag.begin(); i != tag.end(); ++i)
  {
    draw_text(*i, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
    y -= gl_height();
  }
  
  // Playback information.

  x = geom.w - margin;
  y = geom.h - gl_height() - margin;
  str = str::Format("Frame = %1").arg(ui::style::global()->timecode() ?
    image::time::timecode_to_str(
      image::time::frame_to_timecode(_hud_info.playback.frame,
      _hud_info.playback.speed)):
    str::int_to_str(_hud_info.playback.frame));
    
  draw_text(str,
    Vec2i(x - math::ceil(gl_width(str.c_str())), y + gl_descent()),
      _hud_foreground, _hud_background);
  y -= gl_height();

  str = str::Format("%1").arg(_hud_info.playback.speed_real, 0, -2);
  x -= math::ceil(gl_width(str.c_str()));
  draw_text(str, Vec2i(x, y + gl_descent()),
    _hud_info.playback.dropped_frames ? ui::style::global()->color_select() :
    _hud_foreground, _hud_background);
  str = str::Format("Speed = %1/").arg(
    image::time::speed_to_float(_hud_info.playback.speed), 0, -2);
  x -= math::ceil(gl_width(str.c_str()));
  draw_text(str, Vec2i(x, y + gl_descent()), _hud_foreground, _hud_background);
}

void Widget::draw_text(const Str & in, const Vec2i & position,
  const image::Color & foreground, const image::Color & background)
{
  djv::image::Color _background(djv::image::pixel::RGB_F32);
  djv::image::color::convert(background, &_background);
  glColor3f(_background.get_f32(0), _background.get_f32(1),
    _background.get_f32(2));
  gl_draw(in.c_str(), position.x + 1, position.y - 1);
  
  djv::image::Color _foreground(djv::image::pixel::RGB_F32);
  djv::image::color::convert(foreground, &_foreground);
  glColor3f(_foreground.get_f32(0), _foreground.get_f32(1),
    _foreground.get_f32(2));
  gl_draw(in.c_str(), position.x, position.y);
}

const List<Str> & label_grid()
{
  static const List<Str> data = List<Str>() << "None" << "1x1" << "10x10" <<
    "100x100";
  return data;
}

}
}

