/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_fnc.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "gl_data.h"
#include <djv_base/mtx.h>

namespace djv {
namespace image {
namespace gl {
namespace data
{

Mtx4<GLfloat> color_mtx(const Color & in)
{
  const Mtx4<GLfloat> brightness(
    in.brightness, 0, 0, 0,
    0, in.brightness, 0, 0,
    0, 0, in.brightness, 0,
    0, 0, 0, 1);

  const Mtx4<GLfloat> contrast =
    Mtx4<GLfloat>(
      1, 0, 0, 0,
      0, 1, 0, 0,
      0, 0, 1, 0,
      -0.5, -0.5, -0.5, 1) *
    Mtx4<GLfloat>(
      in.contrast, 0, 0, 0,
      0, in.contrast, 0, 0,
      0, 0, in.contrast, 0,
      0, 0, 0, 1) *
    Mtx4<GLfloat>(
      1, 0, 0, 0,
      0, 1, 0, 0,
      0, 0, 1, 0,
      0.5, 0.5, 0.5, 1);

  const float s [] = {
    (1.0 - in.saturation) * 0.3086,
    (1.0 - in.saturation) * 0.6094,
    (1.0 - in.saturation) * 0.0820 };
  const Mtx4<GLfloat> saturation(
    s[0] + in.saturation, s[0], s[0], 0,
    s[1], s[1] + in.saturation, s[1], 0,
    s[2], s[2], s[2] + in.saturation, 0,
    0, 0, 0, 1 );

  return brightness * contrast * saturation;
}

image::Data color_lut(const Levels & in, double soft_clip)
{
  image::Data out(image::data::Info(Vec2i(1024, 1), pixel::L_F32));
  
  const double in_tmp = in.in[1] - in.in[0];
  const double gamma = 1.0 / in.gamma;
  const double out_tmp = in.out[1] - in.out[0];

  pixel::F32_T * p = reinterpret_cast<pixel::F32_T *>(out.data());
  const int size = out.size().x;
  for (int i = 0; i < size; ++i, ++p)
    *p = math::soft_clip(
           (math::pow(math::max(
             ((i / double(size - 1) - in.in[0]) / in_tmp), 0.000001),
             gamma) * out_tmp + in.out[0]),
           soft_clip);
  
  return out;
}

void unpack(const image::Info & in, const Vec2i & offset)
{
  glPixelStorei(GL_UNPACK_ALIGNMENT, in.align);
  glPixelStorei(GL_UNPACK_SWAP_BYTES, in.endian != mem::endian());
  glPixelStorei(GL_UNPACK_ROW_LENGTH, in.size.x);
  glPixelStorei(GL_UNPACK_SKIP_ROWS, offset.y);
  glPixelStorei(GL_UNPACK_SKIP_PIXELS, offset.x);
}

void pack(const image::Info & in, const Vec2i & offset)
{
  glPixelStorei(GL_PACK_ALIGNMENT, in.align);
  glPixelStorei(GL_PACK_SWAP_BYTES, in.endian != mem::endian());
  glPixelStorei(GL_PACK_ROW_LENGTH, in.size.x);
  glPixelStorei(GL_PACK_SKIP_ROWS, offset.y);
  glPixelStorei(GL_PACK_SKIP_PIXELS, offset.x);
}

bool operator == (const Color & a, const Color & b)
{
  return a.brightness == b.brightness && a.contrast == b.contrast &&
    a.saturation == b.saturation;
}

bool operator != (const Color & a, const Color & b)
{ return ! (a == b); }

bool operator == (const Levels & a, const Levels & b)
{
  return a.in[0] == b.in[0] && a.in[1] == b.in[1] && a.gamma == b.gamma &&
    a.out[0] == b.out[0] && a.out[1] == b.out[1];
}

bool operator != (const Levels & a, const Levels & b)
{ return ! (a == b); }

bool operator == (const Display & a, const Display & b)
{
  return a.lut.info() == b.lut.info() && a.color == b.color &&
    a.levels == b.levels && a.soft_clip == b.soft_clip;
}

bool operator != (const Display & a, const Display & b)
{ return ! (a == b); }

bool serialize(List<Str> * in, Color * out)
{
  if (! str::serialize(in, &out->brightness)) return false;
  if (! str::serialize(in, &out->contrast)) return false;
  if (! str::serialize(in, &out->saturation)) return false;
  return true;
}

List<Str> serialize(const Color & in)
{
  return List<Str>() << str::serialize(in.brightness) <<
    str::serialize(in.contrast) << str::serialize(in.saturation);
}

bool serialize(List<Str> * in, Levels * out)
{
  if (! str::serialize(in, &out->in[0])) return false;
  if (! str::serialize(in, &out->in[1])) return false;
  if (! str::serialize(in, &out->gamma)) return false;
  if (! str::serialize(in, &out->out[0])) return false;
  if (! str::serialize(in, &out->out[1])) return false;
  return true;
}

List<Str> serialize(const Levels & in)
{
  return List<Str>() << str::serialize(in.in[0]) << str::serialize(in.in[1]) <<
    str::serialize(in.gamma) << str::serialize(in.out[0]) <<
    str::serialize(in.out[1]);
}

bool serialize(List<Str> * in, Display * out)
{
  if (! serialize(in, &out->color)) return false;
  if (! serialize(in, &out->levels)) return false;
  if (! str::serialize(in, &out->soft_clip)) return false;
  return true;
}

List<Str> serialize(const Display & in)
{
  return List<Str>() << serialize(in.color) << serialize(in.levels) <<
    str::serialize(in.soft_clip);
}

const List<Str> & label_channel()
{
  static const List<Str> data = List<Str>() << "Default" << "0" << "1" <<
    "2" << "3";
  return data;
}

const List<Str> & label_filter()
{
  static const List<Str> data = List<Str>() << "Linear" << "Nearest";
  return data;
}

namespace
{
Vec2i _texture_tile(256);
bool _texture_dbg = false;
}

void texture_tile(const Vec2i & in)
{
  _texture_tile = Vec2i(math::to_pow2(in.x), math::to_pow2(in.y));
}

const Vec2i & texture_tile() { return _texture_tile; }

void texture_dbg(bool in) { _texture_dbg = in; }
bool texture_dbg() { return _texture_dbg; }

Dbg & operator << (Dbg & dbg, const Color & in)
{ return dbg << serialize(in); }
Dbg & operator << (Dbg & dbg, const Levels & in)
{ return dbg << serialize(in); }
Dbg & operator << (Dbg & dbg, const Display & in)
{ return dbg << in.color << " " << in.levels << " " << in.soft_clip; }
Dbg & operator << (Dbg & dbg, CHANNEL in) { return dbg << label_channel()[in]; }
Dbg & operator << (Dbg & dbg, FILTER in) { return dbg << label_filter()[in]; }

}
}
}
}

