/*------------------------------------------------------------------------------
 lib/djv_image/io_cineon_header.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_cineon_header.cxx"

#include "io_cineon.h"

namespace djv {
namespace image {
namespace io {
namespace cineon
{

namespace
{

enum ORIENT
{
  ORIENT_LEFT_RIGHT_TOP_BOTTOM,
  ORIENT_LEFT_RIGHT_BOTTOM_TOP,
  ORIENT_RIGHT_LEFT_TOP_BOTTOM,
  ORIENT_RIGHT_LEFT_BOTTOM_TOP,
  ORIENT_TOP_BOTTOM_LEFT_RIGHT,
  ORIENT_TOP_BOTTOM_RIGHT_LEFT,
  ORIENT_BOTTOM_TOP_LEFT_RIGHT,
  ORIENT_BOTTOM_TOP_RIGHT_LEFT
};

enum DESCRIPTOR
{
  DESCRIPTOR_L = 0,
  DESCRIPTOR_R_PRINT = 1,
  DESCRIPTOR_G_PRINT = 2,
  DESCRIPTOR_B_PRINT = 3,
  DESCRIPTOR_R_CCIR_XA11 = 4,
  DESCRIPTOR_G_CCIR_XA11 = 5,
  DESCRIPTOR_B_CCIR_XA11 = 6
};

struct Header
{
  struct File
  {
    uint32_t magic;
    uint32_t image_offset;
    uint32_t header_size;
    uint32_t industry_header_size;
    uint32_t user_header_size;
    uint32_t size;
    char version [8];
    char name [100];
    char time [24];
    uint8_t _pad [36];
  } file;
  
  struct Image
  {
    uint8_t orient;
    uint8_t channels;
    uint8_t _pad [2];

    struct Channel
    {
      uint8_t descriptor [2];
      uint8_t bit_depth;
      uint8_t _pad;
      uint32_t size [2];
      float low_data;
      float low_quantity;
      float high_data;
      float high_quantity;
    } channel [8];
    
    float white [2];
    float red [2];
    float green [2];
    float blue [2];
    char label [200];
    uint8_t _pad2 [28];
    uint8_t interleave;
    uint8_t packing;
    uint8_t data_sign;
    uint8_t data_sense;
    uint32_t line_padding;
    uint32_t channel_padding;
    uint8_t _pad3 [20];
  } image;
  
  struct Source
  {
    int32_t offset [2];
    char file [100];
    char time [24];
    char input_device [64];
    char input_model [32];
    char input_serial [32];
    float input_pitch [2];
    float gamma;
    char _pad [40];
  } source;
  
  struct Film
  {
    uint8_t id;
    uint8_t type;
    uint8_t offset;
    uint8_t _pad;
    uint32_t prefix;
    uint32_t count;
    char format [32];
    uint32_t frame;
    float frame_rate;
    char frame_id [32];
    char slate [200];
    char _pad2 [740];
  } film;
};

inline void zero(int32_t * in) { *((uint32_t *)in) = 0x80000000; }
inline void zero(float * in) { *((uint32_t *)in) = 0x7F800000; }
inline void zero(char * in, int size) { mem::zero(in, size); }

inline bool valid(const uint8_t * in) { return *in != 0xff; }
inline bool valid(const uint16_t * in) { return *in != 0xffff; }

// XXX Catch uninitialized values.

const int32_t _int_max = 1000000;
const float _float_max = 1000000.0;
const char _char_min = 32;
const char _char_max = 126;

const float _speed_min = 0.000001;

inline bool valid(const uint32_t * in)
{ return *in != 0xffffffff && *in < uint32_t(_int_max); }
inline bool valid(const int32_t * in)
{ return *in != int32_t(0x80000000) && *in > -_int_max && *in < _int_max; }
inline bool valid(const float * in)
{ return *((uint32_t *)in) != 0x7F800000 &&
  *in > -_float_max && *in < _float_max; }

inline bool valid(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p)
    if (*p < _char_min || *p > _char_max)
      return false;
  return size ? (in[0] != 0) : false;
}

inline Str _str(const char * in, int size)
{
  const char * p = in;
  const char * const end = p + size;
  for (; *p && p < end; ++p) ;
  return Str(in, p - in);
}

void _endian(Header * in)
{
  mem::endian(&in->file.image_offset, 1, 4);
  mem::endian(&in->file.header_size, 1, 4);
  mem::endian(&in->file.industry_header_size, 1, 4);
  mem::endian(&in->file.user_header_size, 1, 4);
  mem::endian(&in->file.size, 1, 4);

  for (uint i = 0; i < 8; ++i)
  {
    mem::endian(in->image.channel[i].size, 2, 4);
    mem::endian(&in->image.channel[i].low_data, 1, 4);
    mem::endian(&in->image.channel[i].low_quantity, 1, 4);
    mem::endian(&in->image.channel[i].high_data, 1, 4);
    mem::endian(&in->image.channel[i].high_quantity, 1, 4);
  }
  mem::endian(in->image.white, 2, 4);
  mem::endian(in->image.red, 2, 4);
  mem::endian(in->image.green, 2, 4);
  mem::endian(in->image.blue, 2, 4);
  mem::endian(&in->image.line_padding, 1, 4);
  mem::endian(&in->image.channel_padding, 1, 4);
  
  mem::endian(in->source.offset, 2, 4);
  mem::endian(in->source.input_pitch, 2, 4);
  mem::endian(&in->source.gamma, 1, 4);
  
  mem::endian(&in->film.prefix, 1, 4);
  mem::endian(&in->film.count, 1, 4);
  mem::endian(&in->film.frame, 1, 4);
  mem::endian(&in->film.frame_rate, 1, 4);
}

Str dbg(uint8_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint16_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
Str dbg(uint32_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
//{ return dbg_u32(in); }
Str dbg(int32_t in)
//{ return str::int_to_str(in); }
{ return valid(&in) ? str::int_to_str(in) : "-"; }
//{ return dbg_u32(in); }
Str dbg(float in)
//{ return str::float_to_str(in); }
{ return valid(&in) ? str::float_to_str(in) : "-"; }
//{ return dbg_u32(in); }
Str dbg(const char * in, int size)
//{ return Str(in, size); }
{ return valid(in, size) ? Str(in, size) : "-"; }

void dbg(const Header & in)
{
  DJV_DBG_OBJ2(_DBG, "dbg");
  
  DJV_DBG_PRN("file");
  DJV_DBG_PRN("  image offset = " << dbg(in.file.image_offset));
  DJV_DBG_PRN("  header size = " << dbg(in.file.header_size));
  DJV_DBG_PRN("  industry header size = " << dbg(in.file.industry_header_size));
  DJV_DBG_PRN("  user header size = " << dbg(in.file.user_header_size));
  DJV_DBG_PRN("  size = " << dbg(in.file.size));
  DJV_DBG_PRN("  version = " << dbg(in.file.version, 8));
  DJV_DBG_PRN("  name = " << dbg(in.file.name, 100));
  DJV_DBG_PRN("  time = " << dbg(in.file.time, 24));

  DJV_DBG_PRN("image");
  DJV_DBG_PRN("  orient = " << dbg(in.image.orient));
  DJV_DBG_PRN("  channels = " << dbg(in.image.channels));
  for (int i = 0; i < in.image.channels; ++i)
  {
    DJV_DBG_PRN("  channel " << i);
    DJV_DBG_PRN("    descriptor = " <<
      dbg(in.image.channel[i].descriptor[0]) << " " <<
      dbg(in.image.channel[i].descriptor[1]));
    DJV_DBG_PRN("    bit depth = " << dbg(in.image.channel[i].bit_depth));
    DJV_DBG_PRN("    size = " << dbg(in.image.channel[i].size[0]) << " " <<
      dbg(in.image.channel[i].size[1]));
    DJV_DBG_PRN("    low data = " << dbg(in.image.channel[i].low_data));
    DJV_DBG_PRN("    low quantity = " << dbg(in.image.channel[i].low_quantity));
    DJV_DBG_PRN("    high data = " << dbg(in.image.channel[i].high_data));
    DJV_DBG_PRN("    high quantity = " <<
      dbg(in.image.channel[i].high_quantity));
  }
  DJV_DBG_PRN("  white = " << dbg(in.image.white[0]) << " " <<
    dbg(in.image.white[1]));
  DJV_DBG_PRN("  red = " << dbg(in.image.red[0]) << " " <<
    dbg(in.image.red[1]));
  DJV_DBG_PRN("  green = " << dbg(in.image.green[0]) << " " <<
    dbg(in.image.green[1]));
  DJV_DBG_PRN("  blue = " << dbg(in.image.blue[0]) << " " <<
    dbg(in.image.blue[1]));
  DJV_DBG_PRN("  label = " << dbg(in.image.label, 200));
  DJV_DBG_PRN("  interleave = " << dbg(in.image.interleave));
  DJV_DBG_PRN("  packing = " << dbg(in.image.packing));
  DJV_DBG_PRN("  data sign = " << dbg(in.image.data_sign));
  DJV_DBG_PRN("  data sense = " << dbg(in.image.data_sense));
  DJV_DBG_PRN("  line padding = " << dbg(in.image.line_padding));
  DJV_DBG_PRN("  channel padding = " << dbg(in.image.channel_padding));
  
  DJV_DBG_PRN("source");
  DJV_DBG_PRN("  offset = " << dbg(in.source.offset[0]) << " " <<
    dbg(in.source.offset[1]));
  DJV_DBG_PRN("  file = " << dbg(in.source.file, 100));
  DJV_DBG_PRN("  time = " << dbg(in.source.time, 24));
  DJV_DBG_PRN("  input device = " << dbg(in.source.input_device, 64));
  DJV_DBG_PRN("  input model = " << dbg(in.source.input_model, 32));
  DJV_DBG_PRN("  input serial = " << dbg(in.source.input_serial, 32));
  DJV_DBG_PRN("  input pitch = " << dbg(in.source.input_pitch[0]) << " " <<
    dbg(in.source.input_pitch[1]));
  DJV_DBG_PRN("  gamma = " << dbg(in.source.gamma));

  DJV_DBG_PRN("film");
  DJV_DBG_PRN("  id = " << dbg(in.film.id));
  DJV_DBG_PRN("  type = " << dbg(in.film.type));
  DJV_DBG_PRN("  offset = " << dbg(in.film.offset));
  DJV_DBG_PRN("  prefix = " << dbg(in.film.prefix));
  DJV_DBG_PRN("  count = " << dbg(in.film.count));
  DJV_DBG_PRN("  format = " << dbg(in.film.format, 32));
  DJV_DBG_PRN("  frame = " << dbg(in.film.frame));
  DJV_DBG_PRN("  frame rate = " << dbg(in.film.frame_rate));
  DJV_DBG_PRN("  frame id = " << dbg(in.film.frame_id, 32));
  DJV_DBG_PRN("  slate = " << dbg(in.film.slate, 200));
}

void _set(const Info & info, dpx::COLOR color, Header::File * file,
  Header::Image * image)
{
  zero(file->version, 8);

  file->image_offset = 2048;
  file->header_size = 1024;
  file->industry_header_size = 1024;
  file->user_header_size = 0;
  
  for (uint i = 0; i < 8; ++i)
  {
    zero(&image->channel[i].low_data);
    zero(&image->channel[i].low_quantity);
    zero(&image->channel[i].high_data);
    zero(&image->channel[i].high_quantity);
  }

  image->orient = ORIENT_LEFT_RIGHT_TOP_BOTTOM;

  image->channels = 3;  
  if (dpx::COLOR_PRINT == color)
  {
    image->channel[0].descriptor[1] = DESCRIPTOR_R_PRINT;
    image->channel[1].descriptor[1] = DESCRIPTOR_G_PRINT;
    image->channel[2].descriptor[1] = DESCRIPTOR_B_PRINT;
  }
  else
  {
    image->channel[0].descriptor[1] = DESCRIPTOR_L;
    image->channel[1].descriptor[1] = DESCRIPTOR_L;
    image->channel[2].descriptor[1] = DESCRIPTOR_L;
  }
  
  const int bit_depth = 10;
  for (int i = 0; i < image->channels; ++i)
  {
    image->channel[i].descriptor[0] = 0;
    image->channel[i].bit_depth = bit_depth;
    image->channel[i].size[0] = info.size.x;
    image->channel[i].size[1] = info.size.y;

    image->channel[i].low_data = 0;
    switch (bit_depth)
    {
      case 8: image->channel[i].high_data = 255; break;
      case 10: image->channel[i].high_data = 1023; break;
      case 12: image->channel[i].high_data = 4095; break;
      case 16: image->channel[i].high_data = 65535; break;
    }
  }
  
  image->interleave = 0;
  image->packing = 5;
  image->data_sign = 0;
  image->data_sense = 0;
  image->line_padding = 0;
  image->channel_padding = 0;
}

void _get(const Header::Image & in, Info * info, bool * print) throw (Error)
{
  switch (in.orient)
  {
    case ORIENT_LEFT_RIGHT_TOP_BOTTOM: info->mirror_y = true; break;
    case ORIENT_LEFT_RIGHT_BOTTOM_TOP: break;
    case ORIENT_RIGHT_LEFT_TOP_BOTTOM: info->mirror_x = true; break;
    case ORIENT_RIGHT_LEFT_BOTTOM_TOP:
      info->mirror_x = info->mirror_y = true; break;
    case ORIENT_TOP_BOTTOM_LEFT_RIGHT:
    case ORIENT_TOP_BOTTOM_RIGHT_LEFT:
    case ORIENT_BOTTOM_TOP_LEFT_RIGHT:
    case ORIENT_BOTTOM_TOP_RIGHT_LEFT: break;
  }
  
  if (! in.channels)
    throw Error(name, error_unsupported);
  int i = 1;
  for (; i < in.channels; ++i)
  {
    if ((in.channel[i].size[0] != in.channel[0].size[0]) ||
      (in.channel[i].size[1] != in.channel[0].size[1]))
      break;
    if (in.channel[i].bit_depth != in.channel[0].bit_depth)
      break;
  }
  if (i < in.channels)
    throw Error(name, error_unsupported);
  
  int pixel = -1;
  switch (in.channels)
  {
    case 3:
      switch (in.channel[0].bit_depth)
      {
        case 10: pixel = pixel::RGB_U10; break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(name, error_unsupported);
  info->pixel = PIXEL(pixel);
  
  info->size = Vec2i(in.channel[0].size[0], in.channel[0].size[1]);
      
  if (valid(&in.line_padding) && in.line_padding)
    throw Error(name, error_unsupported);
  if (valid(&in.channel_padding) && in.channel_padding)
    throw Error(name, error_unsupported);
  
  if (print)
    *print = DESCRIPTOR_R_PRINT == in.channel[0].descriptor[1];
}

void _tag(const Str & file_name, const Tag & tag, Header::File * file,
  Header::Source * source, Header::Film * film)
{
  const List<Str> & label = tag::label();
  const List<Str> & cineon_label = cineon::label_tag();
  Str _tag;
  
  // File.
  
  zero(file->name, 100);
  str::ascii(file_name, file->name, 100, false);

  zero(file->time, 24);
  str::ascii(tag[label[tag::TIME]], file->time, 24, false);
  
  // Source.
  
  zero(&source->offset[0]);
  zero(&source->offset[1]);
  _tag = tag[cineon_label[TAG_SOURCE_OFFSET]];
  if (_tag.size())
  {
    const List<Str> tmp = str::split(_tag, " ");
    if (2 == tmp.size())
    {
      source->offset[0] = str::str_to_int(tmp[0]);
      source->offset[1] = str::str_to_int(tmp[1]);
    }
  }
  
  zero(source->file, 100);
  _tag = tag[cineon_label[TAG_SOURCE_FILE]];
  if (_tag.size()) str::ascii(_tag, source->file, 100, false);

  zero(source->time, 24);
  _tag = tag[cineon_label[TAG_SOURCE_TIME]];
  if (_tag.size()) str::ascii(_tag, source->time, 24, false);

  zero(source->input_device, 64);
  _tag = tag[cineon_label[TAG_SOURCE_INPUT_DEVICE]];
  if (_tag.size()) str::ascii(_tag, source->input_device, 64, false);

  zero(source->input_model, 32);
  _tag = tag[cineon_label[TAG_SOURCE_INPUT_MODEL]];
  if (_tag.size()) str::ascii(_tag, source->input_model, 32, false);

  zero(source->input_serial, 32);
  _tag = tag[cineon_label[TAG_SOURCE_INPUT_SERIAL]];
  if (_tag.size()) str::ascii(_tag, source->input_serial, 32, false);
  
  zero(&source->input_pitch[0]);
  zero(&source->input_pitch[1]);
  _tag = tag[cineon_label[TAG_SOURCE_INPUT_PITCH]];
  if (_tag.size())
  {
    const List<Str> tmp = str::split(_tag, " ");
    if (2 == tmp.size())
    {
      source->offset[0] = str::str_to_int(tmp[0]);
      source->offset[1] = str::str_to_int(tmp[1]);
    }
  }
  
  zero(&source->gamma);
  _tag = tag[cineon_label[TAG_SOURCE_GAMMA]];
  if (_tag.size()) source->gamma = str::str_to_float(_tag);
  
  // Film.

  _tag = tag[label[tag::KEYCODE]];
  if (_tag.size())
  {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    time::str_to_keycode(_tag, &id, &type, &prefix, &count, &offset);
    film->id = id;
    film->type = type;
    film->offset = offset;
    film->prefix = prefix;
    film->count = count;
  }
  
  zero(film->format, 32);
  _tag = tag[cineon_label[TAG_FILM_FORMAT]];
  if (_tag.size()) str::ascii(_tag, film->format, 32, false);

  _tag = tag[cineon_label[TAG_FILM_FRAME]];
  if (_tag.size()) film->frame = str::str_to_int(_tag);

  zero(&film->frame_rate);
  _tag = tag[cineon_label[TAG_FILM_FRAME_RATE]];
  if (_tag.size()) film->frame_rate = str::str_to_float(_tag);

  zero(film->frame_id, 32);
  _tag = tag[cineon_label[TAG_FILM_FRAME_ID]];
  if (_tag.size()) str::ascii(_tag, film->frame_id, 32, false);

  zero(film->slate, 200);
  _tag = tag[cineon_label[TAG_FILM_SLATE]];
  if (_tag.size()) str::ascii(_tag, film->slate, 200, false);
}

void _tag(const Header::File & file, const Header::Source & source,
  const Header::Film & film, Tag * tag, time::Speed * speed)
{
  const List<Str> & label = tag::label();
  const List<Str> & cineon_label = cineon::label_tag();
  
  // File.
  
  if (valid(file.time, 24))
    (*tag)[label[tag::TIME]] = _str(file.time, 24);
    
  // Source.
  
  if (valid(&source.offset[0]) && valid(&source.offset[1]))
    (*tag)[cineon_label[TAG_SOURCE_OFFSET]] = str::join(List<Str>() <<
      str::int_to_str(source.offset[0]) <<
      str::int_to_str(source.offset[1]), " ");

  if (valid(source.file, 100))
    (*tag)[cineon_label[TAG_SOURCE_FILE]] = _str(source.file, 100);
  
  if (valid(source.time, 24))
    (*tag)[cineon_label[TAG_SOURCE_TIME]] = _str(source.time, 24);
  
  if (valid(source.input_device, 64))
    (*tag)[cineon_label[TAG_SOURCE_INPUT_DEVICE]] =
      _str(source.input_device, 64);
  
  if (valid(source.input_model, 32))
    (*tag)[cineon_label[TAG_SOURCE_INPUT_MODEL]] =
      _str(source.input_model, 32);
  
  if (valid(source.input_serial, 32))
    (*tag)[cineon_label[TAG_SOURCE_INPUT_SERIAL]] =
      _str(source.input_serial, 32);
  
  if (valid(&source.input_pitch[0]) && valid(&source.input_pitch[1]))
    (*tag)[cineon_label[TAG_SOURCE_INPUT_PITCH]] = str::join(List<Str>() <<
      str::float_to_str(source.input_pitch[0]) <<
      str::float_to_str(source.input_pitch[1]), " ");
   
  if (valid(&source.gamma))
    (*tag)[cineon_label[TAG_SOURCE_GAMMA]] = str::float_to_str(source.gamma);
      
  // Film.
  
  if (valid(&film.id) && valid(&film.type) && valid(&film.offset) &&
    valid(&film.prefix) && valid(&film.count))
    (*tag)[label[tag::KEYCODE]] = time::keycode_to_str(
      film.id, film.type, film.prefix, film.count, film.offset);

  if (valid(film.format, 32))
    (*tag)[cineon_label[TAG_FILM_FORMAT]] = _str(film.format, 32);
  
  if (valid(&film.frame))
    (*tag)[cineon_label[TAG_FILM_FRAME]] = str::int_to_str(film.frame);
  
  if (valid(&film.frame_rate) && film.frame_rate >= _speed_min)
  {
    if (speed) *speed = time::float_to_speed(film.frame_rate);
    (*tag)[cineon_label[TAG_FILM_FRAME_RATE]] =
      str::float_to_str(film.frame_rate);
  }
  
  if (valid(film.frame_id, 32))
    (*tag)[cineon_label[TAG_FILM_FRAME_ID]] = _str(film.frame_id, 32);
  
  if (valid(film.slate, 200))
    (*tag)[cineon_label[TAG_FILM_SLATE]] = _str(film.slate, 200);
}

}

void header_load(djv::Io * io, Info * info, Tag * tag, time::Speed * speed,
  bool * print) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_load");
  
  Header header;
  io->get(&header, sizeof(Header));
  if (0x802a5fd7 == header.file.magic)
  {
    io->endian(false);
  }
  else if (0xd75f2a80 == header.file.magic)
  {
    DJV_DBG_PRN("endian");
    io->endian(true);
    _endian(&header);
    info->endian = mem::endian_opposite(mem::endian());
  }
  else
  {
    throw Error(name, error_unrecognized);
  }
  dbg(header);
  
  _get(header.image, info, print);
  _tag(header.file, header.source, header.film, tag, speed);
  
  if (header.file.image_offset)
    io->position(header.file.image_offset);
}

void header_save(djv::Io * io, const Str & file, const Info & info,
  const Tag & tag, dpx::COLOR color) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "header_save");
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("color = " << color);
  
  Header header;
  mem::set(0xff, &header, sizeof(Header));
  _set(info, color, &header.file, &header.image);
  _tag(file, tag, &header.file, &header.source, &header.film);
  dbg(header);
  
  const bool endian = mem::endian() != mem::MSB;
  io->endian(endian);
  if (endian)
  {
    DJV_DBG_PRN("endian");
    _endian(&header);
    header.file.magic = 0xd75f2a80;
  }
  else
  {
    header.file.magic = 0x802a5fd7;
  }
  io->set(&header, sizeof(Header));
}

void header_save_update(djv::Io * io) throw (Error)
{
  const size_t size = io->position();
  io->position(20);
  io->set_u32(size);
}

}
}
}
}

