/*------------------------------------------------------------------------------
 lib/djv_image/io_dpx_load.cxx
 
 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_dpx_load.cxx"

#include "io_dpx.h"

namespace djv {
namespace image {
namespace io {
namespace dpx
{

Load::Load() : _print(false) {}

Plugin * Load::copy() const
{ Load * out = new Load; out->_options = _options; return out; }

namespace
{
const Str _help =
"* dpx color (value) - Input color space: %1. Default: %2.\n"
"* dpx print (black) (white) (gamma) (soft clip) - Print values. Default: "
"%3.\n";
}

Str Load::help() const
{
  return str::Format(_help).
    arg(label_color(), ", ").
    arg(str::serialize(_options.color, label_color())).
    arg(serialize(_options.print));
}

bool Load::option(const Str & in, List<Str> * data)
{
  _print_lut.reset();
  
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(data, &_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_PRINT]))
    return serialize(data, &_options.print);
  return false;
}

List<Str> Load::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COLOR]))
    return str::serialize(_options.color, label_color());
  else if (str::compare(in, option_list()[OPTION_PRINT]))
    return serialize(_options.print);
  return List<Str>();
}

const List<Str> & Load::option_list() const { return label_load_option(); }

void Load::open(const File & in, Info * info, Tag * tag, Time * time)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info, tag, &time->speed);
  
  if (file::SEQ == _file.type)
  {
    time->start = _file.seq.start;
    time->end = _file.seq.end;
  }
}

void Load::_open(const File & in, Info * info, Tag * tag, time::Speed * speed)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);

  _io.open(in, djv::io::READ);
  
  _print = false;
  header_load(&_io, info, tag, speed, &_print);
  DJV_DBG_PRN("info " << *info);
}

const Image * Load::get(int frame, int, int proxy) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Setup.

  if (dpx::COLOR_PRINT == _options.color ||
    (dpx::COLOR_AUTO == _options.color && _print))
  {
    DJV_DBG_PRN("profile");
    _image.profile.type = profile::LUT;
    if (! _print_lut.get()) _print_lut = std::auto_ptr<Data>(
      new Data(dpx::unprint_lut(_options.print)));
    _image.profile.lut = *_print_lut;
  }
  else
  {
    _image.profile = Profile();
  }
  
  const File file = File(_file.get(frame != -1 ? frame : _file.seq.start));
  DJV_DBG_PRN("file = " << file);
  
  const Cache cache(file, frame, proxy, 0);
  if (cache == _cache)
  {
    DJV_DBG_PRN("cached");
    return &_image;
  }
  else
  {
    _cache = cache;
  }

  // Open.
  
  Info info;
  _open(file, &info, &_image.tag);

  // Read.
  
  const uint8_t * p = _io.mmap_p();
  _io.seek(data::data_bytes(info));
  
  if (! proxy)
  {
    _image.set(info, p);
  }
  else
  {
    _tmp.set(info, p);
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);  
  return &_image;
}

}
}
}
}

