/*------------------------------------------------------------------------------
 lib/djv_image/io_lut_load.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_lut_load.cxx"

#include "io_lut.h"

namespace djv {
namespace image {
namespace io {
namespace lut
{

Plugin * Load::copy() const
{ Load * out = new Load; out->_options = _options; return out; }

namespace
{
const Str _help =
"* lut spec (value) - File specification: %1. Default: %2.\n"
"* lut type (value) - Pixel type: %3. Default: %4.\n";
}

Str Load::help() const
{
  return str::Format(_help).
    arg(label_spec(), ", ").arg(str::serialize(_options.spec, label_spec())).
    arg(label_type(), ", ").arg(str::serialize(_options.type, label_type()));
}

bool Load::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_SPEC]))
    return str::serialize(data, &_options.spec, label_spec());
  else if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(data, &_options.type, label_type());
  return false;
}

List<Str> Load::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_SPEC]))
    return str::serialize(_options.spec, label_spec());
  else if (str::compare(in, option_list()[OPTION_TYPE]))
    return str::serialize(_options.type, label_type());
  return List<Str>();
}

const List<Str> & Load::option_list() const { return label_load_option(); }

void Load::open(const File & in, Info * info, Tag *, Time * time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start), info);

  if (file::SEQ == _file.type)
  {
    time->start = _file.seq.start;
    time->end = _file.seq.end;
  }
}

namespace
{

int _bit_depth(djv::Io * io) throw (Error)
{
  int out = 0;
  const off_t position = io->position();
  char tmp [str::cstr_size];
  while (1)
  {
    try { djv::io::word(io, tmp, str::cstr_size); } catch (Error) { break; }
    out = math::max(str::str_to_int(tmp, str::cstr_size), out);
  }
  io->position(position);
  
  if (out <= pixel::u8_max) return 8;
  else if (out <= pixel::u10_max) return 10;
  return 16;
}

void _open_inferno(djv::Io * io, Info * info, TYPE type) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_open_inferno");
  
  // Header.
  
  char tmp [str::cstr_size];
  djv::io::word(io, tmp, str::cstr_size);
  DJV_DBG_PRN("magic = " << tmp);
  if (Str(tmp) != "LUT:")
    throw Error(name, error_unrecognized);
  
  djv::io::word(io, tmp, str::cstr_size);
  const int channels = str::str_to_int(tmp, str::cstr_size);
  djv::io::word(io, tmp, str::cstr_size);
  const int size = str::str_to_int(tmp, str::cstr_size);
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("channels = " << channels);
  
  // Information.
  
  info->size = Vec2i(size, 1);
  
  int bit_depth = 0;
  switch (type)
  {
    case TYPE_AUTO: bit_depth = _bit_depth(io); break;
    case TYPE_U8: bit_depth = 8; break;
    case TYPE_U10: bit_depth = 10; break;
    case TYPE_U16: bit_depth = 16; break;
  }
  DJV_DBG_PRN("bit depth = " << bit_depth);
  
  if (! pixel::pixel(channels, bit_depth, &info->pixel))
    throw Error(name, error_unsupported);
}

void _open_kodak(djv::Io * io, Info * info, TYPE type) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_open_kodak");
  
  // Header.
  
  const off_t position = io->position();
  Str header;
  int8_t c = 0;
  bool comment = false;
  while (1)
  {
    try { io->get_8(&c); } catch (Error) { break; }
    if ('#' == c) comment = true;
    else if ('\n' == c)
    {
      if (comment) comment = false;
      else break;
    }
    else
    {
      if (! comment) header += c;
    }
  }
  DJV_DBG_PRN("header = " << header);
  
  int channels = str::split(header, " \t").size();
  int size = 1;
  comment = false;
  while (1)
  {
    try { io->get_8(&c); } catch (Error) { break; }
    if ('\n' == c) ++size;
  }
  io->position(position);
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("channels = " << channels);
  
  // Information.
  
  info->size = Vec2i(size, 1);
  
  int bit_depth = 0;
  switch (type)
  {
    case TYPE_AUTO: bit_depth = _bit_depth(io); break;
    case TYPE_U8: bit_depth = 8; break;
    case TYPE_U10: bit_depth = 10; break;
    case TYPE_U16: bit_depth = 16; break;
  }
  DJV_DBG_PRN("bit depth = " << bit_depth);
  
  if (! pixel::pixel(channels, bit_depth, &info->pixel))
    throw Error(name, error_unsupported);
}

}

void Load::_open(const File & in, Info * info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::_open");
  DJV_DBG_PRN("in = " << in);
    
  _io.open(in, djv::io::READ);
  
  _spec = _options.spec;
  if (SPEC_AUTO == _spec)
  {
    if (lut::extension_list[0] == in.extension()) _spec = SPEC_INFERNO;
    else _spec = SPEC_KODAK;
  }
  switch (_spec)
  {
    case SPEC_INFERNO: _open_inferno(&_io, info, _options.type); break;
    case SPEC_KODAK: _open_kodak(&_io, info, _options.type); break;
    default: break;
  }
}

void _read_inferno(djv::Io * io, Image * out) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_read_kodak");
  
  List<Color> color;
  color.resize(out->w());
  for (int x = 0; x < out->w(); ++x)
    color[x].set(out->pixel());
    
  for (int c = 0; c < pixel::channels(out->pixel()); ++c)
    for (int x = 0; x < out->w(); ++x)
    {
      char tmp [str::cstr_size];
      djv::io::word(io, tmp, str::cstr_size);
      const int v = str::str_to_int(tmp, str::cstr_size);
      
      switch (pixel::type(out->pixel()))
      {
        case pixel::U8: color[x].set_u8(v, c); break;
        case pixel::U10: color[x].set_u10(v, c); break;
        case pixel::U16: color[x].set_u16(v, c); break;
        default: break;
      }
    }
  
  for (int x = 0; x < out->w(); ++x)
  {
    DJV_DBG_PRN(x << " = " << color[x]);
    mem::copy(color[x].data(), out->data(x, 0), pixel::bytes(out->pixel()));
  }
}

void _read_kodak(djv::Io * io, Image * out) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_read_kodak");
  
  for (int x = 0; x < out->w(); ++x)
  {
    Color color(out->pixel());
    for (int c = 0; c < pixel::channels(out->pixel()); ++c)
    {
      char tmp [str::cstr_size];
      djv::io::word(io, tmp, str::cstr_size);
      const int v = str::str_to_int(tmp, str::cstr_size);
      
      switch (pixel::type(out->pixel()))
      {
        case pixel::U8: color.set_u8(v, c); break;
        case pixel::U10: color.set_u10(v, c); break;
        case pixel::U16: color.set_u16(v, c); break;
        default: break;
      }
    }
    DJV_DBG_PRN(x << " = " << color);
    mem::copy(color.data(), out->data(x, 0), pixel::bytes(out->pixel()));
  }
}

const Image * Load::get(int frame, int, int) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Load::get");
  DJV_DBG_PRN("frame = " << frame);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start), &info);
  
  // Read.
  
  _image.set(info);
  switch (_spec)
  {
    case SPEC_INFERNO: _read_inferno(&_io, &_image); break;
    case SPEC_KODAK: _read_kodak(&_io, &_image); break;
    default: break;
  }
  _io.close();
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
}
}

