/*------------------------------------------------------------------------------
 lib/djv_image/io_lut_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_lut_save.cxx"

#include "io_lut.h"

namespace djv {
namespace image {
namespace io {
namespace lut
{

Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* lut spec (value) - File specification: %1. Default: %2.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_spec(), ", ").arg(str::serialize(_options.spec, label_spec()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_SPEC]))
    return str::serialize(data, &_options.spec, label_spec());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_SPEC]))
    return str::serialize(_options.spec, label_spec());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &, const Time &)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = Info();
  _info.size = Vec2i(info.size.x, 1);
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type)
  {
    case pixel::F16: 
    case pixel::F32: type = pixel::U16; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

namespace
{

void _open_inferno(djv::Io * io, const Info & info) throw (Error)
{
  char tmp [str::cstr_size];
  int size = SNPRINTF(tmp, str::cstr_size, "LUT: %d %d\n\n",
    pixel::channels(info.pixel), info.size.x);
  io->set(tmp, size);
}

void _open_kodak(djv::Io *, const Info &) throw (Error) {}

}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _io.open(in, djv::io::WRITE);
  
  _spec = _options.spec;
  if (SPEC_AUTO == _spec)
  {
    if (lut::extension_list[0] == in.extension()) _spec = SPEC_INFERNO;
    else _spec = SPEC_KODAK;
  }
  switch (_spec)
  {
    case SPEC_INFERNO: _open_inferno(&_io, _info); break;
    case SPEC_KODAK: _open_kodak(&_io, _info); break;
    default: break;
  }
}

void _write_inferno(djv::Io * io, const Data * out) throw (Error)
{
  List<Color> color;
  color.resize(out->w());
  for (int x = 0; x < out->w(); ++x)
  {
    color[x].set(out->pixel());
    mem::copy(out->data(x, 0), color[x].data(), pixel::bytes(out->pixel()));
  }
  
  for (int c = 0; c < pixel::channels(out->pixel()); ++c)
    for (int x = 0; x < out->w(); ++x)
    {
      int v = 0;
      switch (pixel::type(out->pixel()))
      {
        case pixel::U8: v = color[x].get_u8(c); break;
        case pixel::U10: v = color[x].get_u10(c); break;
        case pixel::U16: v = color[x].get_u16(c); break;
        default: break;
      }
      char tmp [str::cstr_size];
      int size = SNPRINTF(tmp, str::cstr_size, "%9d\n", v);
      io->set(tmp, size);
    }
}

void _write_kodak(djv::Io * io, const Data * out) throw (Error)
{
  for (int x = 0; x < out->w(); ++x)
  {
    Color color(out->pixel());
    mem::copy(out->data(x, 0), color.data(), pixel::bytes(out->pixel()));
    for (int c = 0; c < pixel::channels(out->pixel()); ++c)
    {
      int v = 0;
      switch (pixel::type(out->pixel())) {
        case pixel::U8: v = color.get_u8(c); break;
        case pixel::U10: v = color.get_u10(c); break;
        case pixel::U16: v = color.get_u16(c); break;
        default: break;
      }
      char tmp [str::cstr_size];
      int size = SNPRINTF(tmp, str::cstr_size, "%6d", v);
      io->set(tmp, size);
    }
    io->set_8('\n');
  }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("frame = " << frame);
  
  // Open.

  _open(_file.get(frame));
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    p = &_image;
  }
  
  switch (_spec)
  {
    case SPEC_INFERNO: _write_inferno(&_io, p); break;
    case SPEC_KODAK: _write_kodak(&_io, p); break;
    default: break;
  }
  _io.close();
}

}
}
}
}
