/*------------------------------------------------------------------------------
 lib/djv_image/time.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/time.cxx"

#include "time.h"
#include <djv_base/math.h>

namespace djv {
namespace image {
namespace time
{

Speed::Speed(FPS in) : scale(0), duration(1)
{
  switch (in)
  {
    case FPS_1: scale = 1; break;
    case FPS_3: scale = 3; break;
    case FPS_6: scale = 6; break;
    case FPS_12: scale = 12; break;
    case FPS_15: scale = 15; break;
    case FPS_16: scale = 16; break;
    case FPS_18: scale = 18; break;
    case FPS_23_98: scale = 24000; duration = 1001; break;
    case FPS_24: scale = 24; break;
    case FPS_25: scale = 25; break;
    case FPS_29_97: scale = 30000; duration = 1001; break;
    case FPS_30: scale = 30; break;
    case FPS_50: scale = 50; break;
    case FPS_59_94: scale = 60000; duration = 1001; break;
    case FPS_60: scale = 60; break;
    case FPS_120: scale = 120; break;
  }
}

Time::Time() : start(0), end(0), speed(time::speed()) {}

Time::Time(int start, int end) : start(start), end(end), speed(time::speed())
{}

Time::Time(int start, int end, const Speed & speed) : start(start), end(end),
  speed(speed)
{}

// XXX All timecode/keycode stuff is probably wrong.

void timecode_to_time(uint32_t in, int * hour, int * minute, int * second,
  int * frame)
{
  *hour = (in >> 28 & 0x0f) * 10 + (in >> 24 & 0x0f);
  *minute = (in >> 20 & 0x0f) * 10 + (in >> 16 & 0x0f);
  *second = (in >> 12 & 0x0f) * 10 + (in >>  8 & 0x0f);
  *frame = (in >>  4 & 0x0f) * 10 + (in >>  0 & 0x0f);
}

int timecode_to_frame(uint32_t in, const time::Speed & speed)
{
  if (! speed.valid()) return 0;

  int hour, minute, second, frame;
  timecode_to_time(in, &hour, &minute, &second, &frame);
  
  return int((hour * 60 * 60 + minute * 60 + second) * speed_to_float(speed)) +
    frame;
}

Str timecode_to_str(uint32_t in)
{
  int hour, minute, second, frame;
  timecode_to_time(in, &hour, &minute, &second, &frame);
  
  return str::Format("%1:%2:%3:%4").arg(hour, 2, '0').arg(minute, 2, '0').
    arg(second, 2, '0').arg(frame, 2, '0');
}

uint32_t time_to_timecode(int hour, int minute, int second, int frame)
{
  return
    (hour   / 10 & 0x0f) << 28 | (hour   % 10 & 0x0f) << 24 |
    (minute / 10 & 0x0f) << 20 | (minute % 10 & 0x0f) << 16 |
    (second / 10 & 0x0f) << 12 | (second % 10 & 0x0f) <<  8 |
    (frame  / 10 & 0x0f) <<  4 | (frame  % 10 & 0x0f) <<  0;
}

uint32_t frame_to_timecode(int frame, const time::Speed & speed)
{
  DJV_DBG_OBJ2(_DBG, "frame_to_timecode");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("speed = " << speed);
  
  if (! speed.valid()) return 0;
  const double speed_float = time::speed_to_float(speed);
  
  const int hour = int(frame / (speed_float * 60 * 60));
  frame -= int(hour * speed_float * 60 * 60);
  const int minute = int(frame / (speed_float * 60));
  frame -= int(minute * speed_float * 60);
  const int second = int(frame / speed_float);
  frame -= int(second * speed_float);
  DJV_DBG_PRN("hour = " << hour);
  DJV_DBG_PRN("minute = " << minute);
  DJV_DBG_PRN("second = " << second);
  DJV_DBG_PRN("frame = " << frame);
  
  const uint32_t out = time_to_timecode(hour, minute, second, frame);
  DJV_DBG_PRN("out = " << timecode_to_str(out));
  return out;
}

uint32_t str_to_timecode(const Str & in)
{
  int hour = 0, minute = 0, second = 0, frame = 0;

  List<Str> tmp = str::split(in, ":");
  uint i = 0;
  switch (tmp.size())
  {
    case 4: hour = str::str_to_int(tmp[i++]);
    case 3: minute = str::str_to_int(tmp[i++]);
    case 2: second = str::str_to_int(tmp[i++]);
    case 1: frame = str::str_to_int(tmp[i++]);
  }

  return time_to_timecode(hour, minute, second, frame);
}

Str keycode_to_str(int id, int type, int prefix, int count, int offset)
{
  return str::Format("%1:%2:%3:%4:%5").arg(id).arg(type).arg(prefix).
    arg(count).arg(offset);
}

bool str_to_keycode(const Str & in, int * id, int * type, int * prefix,
  int * count, int * offset)
{
  const List<Str> tmp = str::split(in, ":");
  if (tmp.size() != 5) return false;
  
  *id = str::str_to_int(tmp[0]);
  *type = str::str_to_int(tmp[1]);
  *prefix = str::str_to_int(tmp[2]);
  *count = str::str_to_int(tmp[3]);
  *offset = str::str_to_int(tmp[4]);
  return true;
}

bool operator == (const Speed & a, const Speed & b)
{ return a.scale == b.scale && a.duration == b.duration; }
bool operator != (const Speed & a, const Speed & b)
{ return ! (a == b); }

bool operator == (const Time & a, const Time & b)
{ return a.start == b.start && a.end == b.end && a.speed == b.speed; }
bool operator != (const Time & a, const Time & b)
{ return ! (a == b); }

double speed_to_float(const Speed & in)
{
  return in.scale / double(in.duration);
}

Speed float_to_speed(double in)
{
  if (in == speed_to_float(FPS_1)) return FPS_1;
  else if (in == speed_to_float(FPS_3)) return FPS_3;
  else if (in == speed_to_float(FPS_6)) return FPS_6;
  else if (in == speed_to_float(FPS_12)) return FPS_12;
  else if (in == speed_to_float(FPS_15)) return FPS_15;
  else if (in == speed_to_float(FPS_16)) return FPS_16;
  else if (in == speed_to_float(FPS_18)) return FPS_18;
  else if (in == speed_to_float(FPS_23_98)) return FPS_23_98;
  else if (in == speed_to_float(FPS_24)) return FPS_24;
  else if (in == speed_to_float(FPS_25)) return FPS_25;
  else if (in == speed_to_float(FPS_29_97)) return FPS_29_97;
  else if (in == speed_to_float(FPS_30)) return FPS_30;
  else if (in == speed_to_float(FPS_50)) return FPS_50;
  else if (in == speed_to_float(FPS_59_94)) return FPS_59_94;
  else if (in == speed_to_float(FPS_60)) return FPS_60;
  else if (in == speed_to_float(FPS_120)) return FPS_120;
  return Speed();
}

bool serialize(List<Str> * in, Speed * out)
{
  if (! str::serialize(in, &out->scale)) return false;
  if (! str::serialize(in, &out->duration)) return false;
  return true;
}

bool serialize(List<Str> * in, Time * out)
{
  if (! str::serialize(in, &out->start)) return false;
  if (! str::serialize(in, &out->end)) return false;
  if (! serialize(in, &out->speed)) return false;
  return true;
}

List<Str> serialize(const Speed & in)
{
  return List<Str>() << str::serialize(in.scale) << str::serialize(in.duration);
}

List<Str> serialize(const Time & in)
{
  return List<Str>() << str::serialize(in.start) << str::serialize(in.end) <<
    serialize(in.speed);
}

namespace { Speed _speed(FPS_24); }
void speed(const Speed & in) { _speed = in; }
const Speed & speed() { return _speed; }

const List<Str> & label_fps()
{
  static const List<Str> data = List<Str>() << "1" << "3" << "6" << "12" <<
    "15" << "16" << "18" << "23.98" << "24" << "25" << "29.97" << "30" <<
    "50" << "59.94" << "60" << "120";
  return data;
}

Dbg & operator << (Dbg & dbg, const Speed & in)
{ return dbg << speed_to_float(in); }

Dbg & operator << (Dbg & dbg, const Time & in)
{ return dbg << str::Format("%1-%2:%3").arg(in.start).arg(in.end).
  arg(speed_to_float(in.speed)); }

}
}
}

