/*------------------------------------------------------------------------------
 lib/djv_ui/button_tool.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "button.h"
#include <djv_ui/style.h>
#include <djv_ui/icon.h>
#include <FL/Fl.H>
#include <FL/Fl_Image.H>
#include <FL/fl_draw.H>

namespace djv {
namespace ui {
namespace button
{

Tool::Tool(const Str & icon, int margin) :
  _value(false),
  _type(PUSH),
  _repeat_active(false),
  _icon(ui::icon::get(icon)),
  _margin(margin),
  _shortcut(0),
  _pressed(false), _pressed_inside(false)
{
  if (_icon)
  {
    _icon_inactive = std::auto_ptr<Fl_Image>(
      _icon->copy(_icon->w(), _icon->h()));
    _icon_inactive->inactive();
  }
  
  Item::size(layout::FIXED, layout::FIXED);
}

void Tool::set(bool in)
{
  if (in == _value) return;
  switch (_type)
  {
    case TOGGLE:
      _value = in;
      redraw();
      sig.emit(_value);
      break;
      
    case RADIO:
      _value = in;
      redraw();
      if (_value) {
        sig.emit(_value);
        for (uint i = 0; i < _radio_group.size(); ++i)
          if (_radio_group[i] != this) _radio_group[i]->set(false);
      }
      break;

    default: break;
  }
}

void Tool::type(TYPE in)
{
  _type = in;
}

void Tool::radio_group(const List<Tool *> & in)
{
  _radio_group = in;
}

void Tool::icon(Fl_Image * in)
{
  _icon = in;
  
  if (_icon)
  {
    _icon_inactive = std::auto_ptr<Fl_Image>(
      _icon->copy(_icon->w(), _icon->h()));
    _icon_inactive->inactive();
  }
  else
  {
    _icon_inactive.reset();
  }
}

void Tool::shortcut(int in)
{
  _shortcut = in;
}

int Tool::handle(int in)
{
  const Box2i & geom = this->geom();
  const Vec2i mouse(Fl::event_x(), Fl::event_y());
  switch (in)
  {
    case FL_PUSH:
    {
      _pressed = true;
      _pressed_inside = true;
      if (REPEAT == _type)
        Fl::add_timeout(0.5, repeat_callback, this);
      redraw();
    } return 1;
      
    case FL_RELEASE:
    {
      if (! _repeat_active && intersect(geom, mouse))
        action();
      _repeat_active = false;
      _pressed = false;
      _pressed_inside = false;
      Fl::remove_timeout(repeat_callback, this);
      redraw();
    } return 1;
    
    case FL_DRAG:
      if (REPEAT == _type) break;
      if (intersect(geom, mouse))
      {
        if (! _pressed_inside)
        {
          _pressed_inside = true;
          redraw();
        }
      }
      else
      {
        if (_pressed_inside)
        {
          _pressed_inside = false;
          redraw();
        }
      }
      return 1;
    
    case FL_SHORTCUT:
      if (Fl::event_key() == _shortcut)
      {
        action();
        redraw();
        return 1;
      }
      break;
  }
  return widget::Widget::handle(in);
}

void Tool::draw()
{
  const Box2i & geom = this->geom();
  
  bool down = false;
  if (_value)
  {
    if (_pressed && _pressed_inside) ;
    else down = true;
  }
  else
  {
    if (_pressed && _pressed_inside) down = true;
    else ;
  }

  fl_color(FL_BACKGROUND_COLOR);
  fl_rectf(geom.x, geom.y, geom.w, geom.h);
  if (down)
    fl_draw_box(FL_THIN_DOWN_FRAME, geom.x, geom.y, geom.w, geom.h,
      FL_BACKGROUND_COLOR);
  
  const int x = geom.x + (geom.w - _icon->w()) / 2;
  const int y = geom.y + (geom.h - _icon->h()) / 2;
  
  if (active() && _icon)
    _icon->draw(x, y);
  else if (_icon_inactive.get())
    _icon_inactive->draw(x, y);
}

void Tool::widget_del()
{
  Fl::remove_timeout(repeat_callback, this);
  widget::Widget::widget_del();
}

void Tool::margin(int in)
{
  _margin = in;
}

void Tool::dirty()
{
  widget::Widget::dirty();

  size_hint((_icon ? Vec2i(_icon->w(), _icon->h()) :
    Vec2i(style::global()->size_icon())) + _margin * 2);
}

void Tool::repeat_callback()
{
  if (_pressed_inside)
  {
    _repeat_active = true;
    sig.emit(true);
    Fl::repeat_timeout(0.1, repeat_callback, this);
  }
  else
  {
    _repeat_active = false;
    Fl::remove_timeout(repeat_callback, this);
  }
}

void Tool::action()
{
  switch (_type)
  {
    case PUSH:
    case REPEAT:
      sig.emit(true);
      break;
        
    case TOGGLE:
      _value = ! _value;
      sig.emit(_value);
      break;
        
    case RADIO:
      if (! _value)
      {
        _value = true;
        sig.emit(_value);
        for (uint i = 0; i < _radio_group.size(); ++i)
          if (_radio_group[i] != this) _radio_group[i]->set(false);
      }
      break;
  }
}

}
}
}
