/*------------------------------------------------------------------------------
 lib/djv_ui/layout_row.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
//#undef DJV_DBG
#define _DBG "lib/djv_ui/layout_row.cxx"

#include "layout_row.h"
#include <djv_ui/style.h>

namespace djv {
namespace ui {
namespace layout
{

class _Spacer : public Spacer
{
public:

  _Spacer(const Vec2i & in) : _size(in) { size(FIXED, FIXED); }
  
  void dirty()
  {
    Spacer::dirty();
    
    size_hint(Vec2i((_size.x != -1) ? _size.x : style::global()->spacing(),
      (_size.y != -1) ? _size.y : style::global()->spacing()));
  }

private:

  Vec2i _size;
};
  
class _Stretch : public Spacer {};

Row_Base::Row_Base(Item * parent) : Layout(parent), _stretch(-1) {}

Row_Base::~Row_Base() {}

void Row_Base::add_spacer(const Vec2i & in, bool stretch)
{
  add(new _Spacer(in));
  if (stretch) this->stretch(int(list().size()) - 1);
}

void Row_Base::add_stretch(bool stretch)
{
  add(new _Stretch);
  if (stretch) this->stretch(int(list().size()) - 1);
}

void Row_Base::stretch(int in)
{
  _stretch = in;
}

void Row_Base::stretch(const Item * in)
{
  const List<Item *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i)
    if (in == list[i])
    {
      stretch(i);
      break;
    }
}

void Row_Base::layout(const List<Item *> & in, List<Box2i> * out,
  const Vec2i & size, DIR dir, int spacing, int stretch)
{
  DJV_DBG_OBJ2(_DBG, "Row_Base::layout");
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("direction = " << dir);
  DJV_DBG_PRN("stretch = " << stretch);
  
  const uint list_size = in.size();
  out->resize(list_size);
  
  // Calculate minimum sizes.

  int total_length = 0;
  int stretch_count = 0;
  const Item * prev = 0;

  List<Item *>::const_iterator i, end = in.end();
  List<Box2i>::iterator j;
  int k;

  for (i = in.begin(), j = out->begin(), k = 0; i != end; ++i, ++j, ++k)
  {
    const Item * item = *i;
    
    j->zero();
    
    if (item->visible())
    {
      j->size = item->size_min();
      
      List<Item *>::const_iterator _i = item->group().begin();
      for (const List<Item *>::const_iterator _end = item->group().end();
        _i != _end; ++_i)
      {
        if ((*_i) != item && (*_i)->visible())
          j->size = vec::max((*_i)->size_min(), j->size);
      }

      total_length += j->size.e[dir];

      if (item->item() != SPACER && prev)
        if (prev->item() != SPACER)
          total_length += spacing;

      if (STRETCH == item->size()[dir] || int(k) == stretch)
        ++stretch_count;

      prev = item;
    }

    DJV_DBG_PRN(item->dbg_name() << " = " << *j <<
      " group = " << item->group().size());
  }
  DJV_DBG_PRN("total length = " << total_length);
  
  // Expand to size hints.
  
  for (i = in.begin(), j = out->begin();
    i != end && (size.e[dir] - total_length) > 0; ++i, ++j)
  {
    const Item * item = *i;
    if (item->visible())
    {
      const int tmp = j->size.e[dir];
      
      j->size.e[dir] = math::min(
        math::max(item->size_hint().e[dir], j->size.e[dir]),
        tmp + size.e[dir] - total_length);

      total_length += j->size.e[dir] - tmp;

      DJV_DBG_PRN(item->dbg_name() << " = " << *j);
    }
  }
  DJV_DBG_PRN("total size = " << total_length);
  
  const int stretch_size = math::max(size.e[dir] - total_length, 0);
  DJV_DBG_PRN("stretch count = " << stretch_count);
  DJV_DBG_PRN("stretch size = " << stretch_size);
  
  // Layout items.

  prev = 0;
  int position = 0;
  for (i = in.begin(), j = out->begin(), k = 0; i != end; ++i, ++j, ++k)
  {
    const Item * item = *i;
    
    if (item->visible())
    {
      if (item->item() != SPACER && prev)
        if (prev->item() != SPACER)
          position += spacing;
      
      if (STRETCH == item->size()[dir] || int(k) == stretch)
      {
        int tmp = 0;
        if (int(k) == stretch)
          tmp = stretch_size;
        else if (-1 == stretch && stretch_count)
          tmp = stretch_size / stretch_count;
        j->size.e[dir] += tmp;
      }

      prev = item;
    }
    
    j->position.e[dir] = position;
    position += j->size.e[dir];
    DJV_DBG_PRN(item->dbg_name() << " = " << *j);
  }
}

Horizontal::Horizontal(Item * parent) : Row_Base(parent)
{
  //size(STRETCH, FIXED);
}

void Horizontal::add_spacer(int in, bool stretch)
{
  Row_Base::add_spacer(Vec2i(in, 0), stretch);
}

void Horizontal::dirty()
{
  DJV_DBG_OBJ2(_DBG, "Horizontal::dirty");
  
  Row_Base::dirty();

  layout(list(), &_list, Vec2i(), HORIZONTAL, spacing(), stretch());
  
  int w = 0, h = 0;
  if (_list.size())
    w = _list[_list.size() - 1].x1() + 1 + margin() * 2;
  for (uint i = 0; i < _list.size(); ++i)
    h = math::max(h, _list[i].h);

  _geom_prev.zero();
  size_hint(Vec2i(w, h + margin() * 2));
}

void Horizontal::geom(const Box2i & in)
{
  if (in == _geom_prev) return;
  _geom_prev = in;

  DJV_DBG_OBJ2(_DBG, "Horizontal::geom");
  DJV_DBG_PRN("in = " << in);

  Row_Base::geom(in);

  const List<Item *> & list = this->list();
  
  layout(list, &_list, in.size - margin() * 2, HORIZONTAL, spacing(),
    stretch());
  
  Box2i geom;
  List<Item *>::const_iterator i = list.begin();
  List<Box2i>::const_iterator j = _list.begin();
  for (const List<Item *>::const_iterator end = list.end(); i != end; ++i, ++j)
  {
    Item * item = *i;
    DJV_DBG_PRN(item->dbg_name() << " = " << *j);
    
    int _y = j->y, _h = j->h;
    if (STRETCH == item->size()[VERTICAL])
      _h = math::max(in.h - margin() * 2, _h);
    else
      switch (item->align()[VERTICAL])
      {
        case LEFT: break;
        case CENTER: _y += ((in.h - margin() * 2) - _h) / 2; break;
        case RIGHT: break;
      }
    geom.x = in.x + margin() + j->x;
    geom.y = in.y + margin() + _y;
    geom.w = j->w;
    geom.h = _h;
    item->geom(geom);
  }
}

Vertical::Vertical(Item * parent) : Row_Base(parent)
{
  //size(FIXED, STRETCH);
}

void Vertical::add_spacer(int in, bool stretch)
{
  Row_Base::add_spacer(Vec2i(0, in), stretch);
}

void Vertical::dirty()
{
  DJV_DBG_OBJ2(_DBG, "Vertical::dirty");
  
  Row_Base::dirty();

  const List<Item *> & list = this->list();
  layout(list, &_list, Vec2i(), VERTICAL, spacing(), stretch());
  
  int w = 0, h = 0;
  for (uint i = 0; i < _list.size(); ++i)
    w = math::max(w, _list[i].w);
  if (_list.size())
    h = _list[_list.size() - 1].y1() + 1 + margin() * 2;
  
  _geom_prev.zero();
  size_hint(Vec2i(w + margin() * 2, h));
}

void Vertical::geom(const Box2i & in)
{
  if (in == _geom_prev) return;
  _geom_prev = in;
  
  DJV_DBG_OBJ2(_DBG, "Vertical::geom");
  DJV_DBG_PRN("in = " << in);

  Row_Base::geom(in);
  
  const List<Item *> & list = this->list();
  layout(list, &_list, in.size - margin() * 2, VERTICAL, spacing(), stretch());
  
  Box2i geom;
  List<Item *>::const_iterator i = list.begin();
  List<Box2i>::const_iterator j = _list.begin();
  for (const List<Item *>::const_iterator end = list.end(); i != end; ++i, ++j)
  {
    Item * item = *i;
    DJV_DBG_PRN(item->dbg_name() << " = " << *j);
    
    int _x = j->x, _w = j->w;
    if (STRETCH == item->size()[HORIZONTAL])
      _w = math::max(in.w - margin() * 2, _w);
    else
      switch (item->align()[HORIZONTAL])
      {
        case LEFT: break;
        case CENTER: _x += ((in.w - margin() * 2) - _w) / 2; break;
        case RIGHT: break;
      }
    geom.x = in.x + margin() + _x;
    geom.y = in.y + margin() + j->y;
    geom.w = _w;
    geom.h = j->h;
    item->geom(geom);
  }
}

}
}
}
