/*------------------------------------------------------------------------------
 plugin/djv_jpeg/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_jpeg/save.cxx"

#include "plugin.h"

using libjpeg::jpeg_compress_struct;

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0), _jpeg_init(false) {}
Save::~Save() { _close(); }

plugin::Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* jpeg quality (value) - Compression quality: 1-100. Default: %1.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(_options.quality);
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(data, &_options.quality);
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(_options.quality);
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const image::Info & info, const image::Tag &,
  const image::Time &) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info =image::Info();
  _info.size = info.size;
  image::pixel::FORMAT format = image::pixel::format(info.pixel);
  switch (format)
  {
    case image::pixel::LA: format = image::pixel::L; break;
    case image::pixel::RGBA: format = image::pixel::RGB; break;
    default: break;
  }
  _info.pixel = image::pixel::pixel(format, image::pixel::U8);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

namespace
{
bool _jpeg_init(jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (setjmp(error->jump)) return false;
  libjpeg::jpeg_create_compress(jpeg);
  return true;
}

bool _jpeg_open(FILE * f, jpeg_compress_struct * jpeg, const image::Info & info,
  const image::Tag & tag, int quality, Jpeg_Error * error)
{
  if (setjmp(error->jump)) return false;

  libjpeg::jpeg_stdio_dest(jpeg, f);
  
  jpeg->image_width = info.size.x;
  jpeg->image_height = info.size.y;
  
  if (image::pixel::L_U8 == info.pixel)
  {
    jpeg->input_components = 1;
    jpeg->in_color_space = libjpeg::JCS_GRAYSCALE;
  }
  else if (image::pixel::RGB_U8 == info.pixel)
  {
    jpeg->input_components = 3;
    jpeg->in_color_space = libjpeg::JCS_RGB;
  }
  
  libjpeg::jpeg_set_defaults(jpeg);
  libjpeg::jpeg_set_quality(jpeg, quality, TRUE);
  libjpeg::jpeg_start_compress(jpeg, TRUE);
  
  Str _tag = tag[image::tag::label()[image::tag::DESCRIPTION]];
  if (_tag.size())
    libjpeg::jpeg_write_marker(
      jpeg, JPEG_COM, (libjpeg::JOCTET *)_tag.c_str(), _tag.size());
  
  return true;
}
}

void Save::_open(const File & in, const image::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);
  
  _close();
  
  // Initialize libjpeg.
  
  _jpeg.err = libjpeg::jpeg_std_error(&_jpeg_error.pub);
  _jpeg_error.pub.error_exit = djv_jpeg_error;
  _jpeg_error.pub.emit_message = djv_jpeg_warn;
  _jpeg_error.msg[0] = 0;
  
  if (! ::_jpeg_init(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
  
  _jpeg_init = true;
  
  // Open.
  
  _f = ::fopen(in.get().c_str(), "wb");
  if (! _f)
    throw Error(name(), image::io::error_open);
  
  if (! _jpeg_open(_f, &_jpeg, _info, tag, _options.quality, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
}

void Save::_close()
{
  if (_jpeg_init)
  {
    libjpeg::jpeg_destroy_compress(&_jpeg);
    _jpeg_init = false;
  }
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace
{
bool _scanline(libjpeg::jpeg_compress_struct * jpeg, const uint8_t * in,
  Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::JSAMPROW p [] = { (libjpeg::JSAMPLE *)in };
  if (! libjpeg::jpeg_write_scanlines(jpeg, p, 1)) return false;
  return true;
}

bool _end(libjpeg::jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_finish_compress(jpeg);
  return true;
}
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  _open(_file.get(frame), in.tag);
  
  // Write.
  
  const image::Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    image::gl::data::copy(in, &_image);
    p = &_image;
  }
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (! _scanline(&_jpeg, p->data(0, h - 1 - y), &_jpeg_error))
      throw Error(name(), _jpeg_error.msg);
  
  if (! _end(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);

  _close();
}

