/*------------------------------------------------------------------------------
 plugin/djv_png/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_png/save.cxx"

#include "plugin.h"

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0), _png(0), _png_info(0) {}
Save::~Save() { _close(); }

void Save::open(const File & in, const image::Info & info, const image::Tag &,
  const image::Time &) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  _info = image::Info();
  _info.size = info.size;
  image::pixel::TYPE type = image::pixel::type(info.pixel);
  switch (type)
  {
    case image::pixel::U10:
    case image::pixel::F16: 
    case image::pixel::F32: type = image::pixel::U16; break;
    default: break;
  }
  _info.pixel = image::pixel::pixel(image::pixel::format(info.pixel), type);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

namespace
{
bool _png_open(FILE * f, png_structp png, png_infop * png_info,
  const image::Info & info, const image::Tag & tag)
{
  if (setjmp(png_jmpbuf(png))) return false;

  *png_info = png_create_info_struct(png);
  if (! *png_info) return false;
  
  png_init_io(png, f);
  
  int color_type = 0;
  switch (info.pixel)
  {
    case image::pixel::L_U8:
    case image::pixel::L_U16: color_type = PNG_COLOR_TYPE_GRAY; break;
    case image::pixel::LA_U8:
    case image::pixel::LA_U16: color_type = PNG_COLOR_TYPE_GRAY_ALPHA; break;
    case image::pixel::RGB_U8:
    case image::pixel::RGB_U16: color_type = PNG_COLOR_TYPE_RGB; break;
    case image::pixel::RGBA_U8:
    case image::pixel::RGBA_U16: color_type = PNG_COLOR_TYPE_RGB_ALPHA; break;
    default: break;
  }
  
  png_set_IHDR(png, *png_info,
    info.size.x, info.size.y, image::pixel::bit_depth(info.pixel), color_type,
    PNG_INTERLACE_NONE,
    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
  
  png_write_info(png, *png_info);
  
  return true;
}
}

void Save::_open(const File & in, const image::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Initialize libpng.
  
  _png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, djv_png_error,
    djv_png_warning);
  if (! _png)
    throw Error(name(), _png_error.msg);
  _png->error_ptr = &_png_error;
  
  // Open.
  
  _f = ::fopen(in.get().c_str(), "wb");
  if (! _f)
    throw Error(name(), image::io::error_open);
  if (! _png_open(_f, _png, &_png_info, _image.info(), tag))
    throw Error(name(), _png_error.msg);
}

void Save::_close()
{
  if (_png || _png_info)
  {
    png_destroy_write_struct(
      _png ? &_png : 0,
      _png_info ? &_png_info : 0);
    _png = 0;
    _png_info = 0;
  }
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace
{
bool _scanline(png_structp png, const uint8_t * in)
{
  if (setjmp(png_jmpbuf(png))) return false;
  png_write_row(png, (png_byte *)in);
  return true;
}

bool _end(png_structp png, png_infop png_info)
{
  if (setjmp(png_jmpbuf(png))) return false;
  png_write_end(png, png_info);
  return true;
}
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open file.
  
  _open(_file.get(frame), in.tag);
  
  // Temporary buffers.
  
  const image::Data * p = &in;
  if (in.info() != _image.info())
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    image::gl::data::copy(in, &_image);
    p = &_image;
  }
  
  // Write.
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (! _scanline(_png, p->data(0, h - 1 - y)))
      throw Error(name(), _png_error.msg);

  if (! _end(_png, _png_info))
    throw Error(name(), _png_error.msg);

  _close();
}

