/*------------------------------------------------------------------------------
 plugin/djv_quicktime_apple/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_quicktime_apple/save.cxx"

#include "plugin.h"
#include <QTML.h>

extern "C"
{
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() throw (Error) :
  _frame_duration(0),
  _f(0), _movie(0), _track(0), _media(0), _gworld(0)
{
  DJV_DBG_OBJ2(_DBG, "Save::Save");
  
  quicktime_init();
}

Save::~Save()
{
  DJV_DBG_OBJ2(_DBG, "Save::~Save");
  
  quicktime_del();
}

plugin::Plugin * Save::copy() const
{ Save * out = new Save; out->_options = _options; return out; }

namespace
{
const Str _help =
"* quicktime codec (value) - Codec: %1. Default: %2.\n"
"* quicktime quality (value) - Quality: %3. Default: %4.\n";
}

Str Save::help() const
{
  return str::Format(_help).
    arg(label_codec(), ", ").
    arg(str::serialize(_options.codec, label_codec())).
    arg(label_quality(), ", ").
    arg(str::serialize(_options.quality, label_quality()));
}

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_CODEC]))
    return str::serialize(data, &_options.codec, label_codec());
  else if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(data, &_options.quality, label_quality());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_CODEC]))
    return str::serialize(_options.codec, label_codec());
  else if (str::compare(in, option_list()[OPTION_QUALITY]))
    return str::serialize(_options.quality, label_quality());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const image::Info & info, const image::Tag &,
  const image::Time & time) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("time = " << time);
  
  try { _open(in, info); close(); } catch (Error) {} // XXX Der da der!
  _open(in, info);
  
  _track = ::NewMovieTrack(_movie, ::FixRatio(info.size.x, 1),
    ::FixRatio(info.size.y, 1), kNoVolume);
  if (! _track)
    throw Error(::name, str::Format(error_track).arg(::GetMoviesError()));
  
  _media = ::NewTrackMedia(_track, ::VideoMediaType, time.speed.scale, 0, 0);
  if (! _media)
    throw Error(::name, str::Format(error_media).arg(::GetMoviesError()));
  
  OSErr err = ::BeginMediaEdits(_media);
  if (err != noErr)
    throw Error(::name, str::Format(error_media_begin).arg(err));
  
  // Information.

  _frame_duration = time.speed.duration;

  // Temporary buffers.
  
  image::Info _info;
  _info.size = info.size;
  _info.mirror_y = true;
  _info.pixel = image::pixel::RGBA_U8;
  _image.set(_info);
  
  const ::Rect rect = { 0, 0, _info.size.y, _info.size.x };
  err = ::NewGWorldFromPtr(&_gworld, k32RGBAPixelFormat, &rect, 0, 0, 0,
    (char *)_image.data(0, 0), _info.size.x * 4);
  if (err != noErr)
    throw Error(::name, str::Format(error_gworld).arg(err));
  ::SetMovieGWorld(_movie, _gworld, 0);
}

void Save::_open(const File & in, const image::Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  
  //Str255 file_name;
  //FSSpec file_spec;
  //c2pstrcpy(file_name, file::path_absolute(in.get()).c_str());
  //FSMakeFSSpec(0, 0, file_name, &file_spec);
  
  char file_name [str::cstr_size];
  str::ascii(file::path_absolute(in), file_name, str::cstr_size);
  ::FSSpec file_spec;
  ::NativePathNameToFSSpec(file_name, &file_spec, 0);
  
  OSErr err = ::CreateMovieFile(
    &file_spec, FOUR_CHAR_CODE('TVOD'), ::smCurrentScript,
    ::createMovieFileDeleteCurFile | ::createMovieFileDontCreateResFile,
    &_f, &_movie);
  if (err != noErr)
    throw Error(::name, str::Format(error_open).arg(err));
}

void Save::close() throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::close");

  if (_track && _media)
  {
    DJV_DBG_PRN("end");
    
    OSErr err = ::EndMediaEdits(_media);
    if (err != noErr)
      throw Error(::name, str::Format(error_media_end).arg(err));
    
    err = ::InsertMediaIntoTrack(_track, 0, 0, ::GetMediaDuration(_media),
      fixed1);
    if (err != noErr)
      throw Error(::name, str::Format(error_media_add).arg(err));
    
    short id = ::movieInDataForkResID;
    err = ::AddMovieResource(_movie, _f, &id, 0);
    if (err != noErr)
      throw Error(::name, str::Format(error_movie_add).arg(err));
    
    //FlattenMovieData(
    //  _movie, flattenAddMovieToDataFork,
    //    //| flattenForceMovieResourceBeforeMovieData,
    //  &file_spec, FOUR_CHAR_CODE('TVOD'), smCurrentScript,
    //  createMovieFileDeleteCurFile | createMovieFileDontCreateResFile);
  }
  _track = 0;
  _media = 0;
  
  if (_f) { ::CloseMovieFile(_f); _f = 0; }
  if (_movie) { ::DisposeMovie(_movie); _movie = 0; }
  if (_gworld) { ::DisposeGWorld(_gworld); _gworld = 0; }
}

void Save::set(const image::Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Temporary buffers.
  
  image::gl::data::copy(in, &_image);
  
  // Write.

  const Rect rect = { 0, 0, _image.h(), _image.w() };
  long compress_max = 0;
  OSErr err = ::GetMaxCompressionSize(_gworld->portPixMap, &rect, 0,
    quicktime_quality(_options.quality), quicktime_codec(_options.codec),
    (::CompressorComponent)anyCodec,
    &compress_max);
  if (err != noErr)
    throw Error(::name, str::Format(error_compress).arg(err));
  DJV_DBG_PRN("compress max = " << int(compress_max));

  class _Handle
  {
  public:
    _Handle (long in) throw (Error) : _p(::NewHandle(in))
      { if (! _p) throw Error(::name, error_handle); }
    ~_Handle () { if (_p) ::DisposeHandle(_p); }
    ::Handle operator () () { return _p; }
  private:
    ::Handle _p;
  } compress(compress_max), description(4);
  
  class _Lock_Pixels
  {
  public:
    _Lock_Pixels(PixMapHandle in) : _p(in) { ::LockPixels(_p); }
    ~_Lock_Pixels() { ::UnlockPixels(_p); }
  private:
    ::PixMapHandle _p;
  } lock_pixels(_gworld->portPixMap);
  
  err = ::CompressImage(_gworld->portPixMap, &rect,
    quicktime_quality(_options.quality), quicktime_codec(_options.codec),
    (::ImageDescriptionHandle)description(), *compress());
  if (err != noErr)
    throw Error(::name, str::Format(error_compress).arg(err));
  
  err = ::AddMediaSample(_media, compress(), 0,
    ((::ImageDescription *)*description())->dataSize, _frame_duration,
    (::SampleDescriptionHandle)description(), 1, 0, 0);
  if (err != noErr)
    throw Error(::name, str::Format(error_media_sample).arg(err));
}

